"""Implements theorem matching functions for the Deductive Database (DD)."""

from __future__ import annotations

import collections
import itertools
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    Generator,
    Iterator,
    Optional,
    TypeAlias,
    cast,
)

from newclid.justifications.justification import (
    Justification,
    RuleApplication,
    justify_dependency,
)
from newclid.predicate_types import PredicateArgument
from newclid.predicates import (
    SYMBOLIC_PREDICATES,
    Predicate,
    predicate_from_construction,
)
from newclid.predicates._index import PredicateType
from newclid.predicates.collinearity import Coll
from newclid.predicates.cyclic import Cyclic
from newclid.problem import PredicateConstruction
from newclid.proof_state import ProofState
from newclid.rule import Rule, RuleConstruction, VarName
from newclid.rule_matching.efficient_statement import (
    EfficientStatement,
    efficient_version,
    generate_permutations,
)
from newclid.rule_matching.interface import RuleMatcher
from newclid.rule_matching.permutations import generate_permutations_as_dicts

LOGGER = logging.getLogger(__name__)

Mapping: TypeAlias = dict[VarName, PredicateArgument]


class MappingMatcher(RuleMatcher):
    def __init__(self, theorem_mapper: TheoremMapper) -> None:
        self.runtime_cache_path: Optional[Path] = None
        self.theorem_mapper = theorem_mapper

    def match_theorem(self, rule: Rule, proof: ProofState) -> set[Justification]:
        conclusions = self._match_generic(rule, proof)
        valid_conclusions: set[Justification] = set()
        for conclusion_dep in conclusions:
            if proof.graph.hyper_graph.get(conclusion_dep.predicate) is not None:
                continue
            applicable = True
            for premise in justify_dependency(conclusion_dep, proof):
                if not proof.check(premise):
                    applicable = False
            if not applicable and "Pythagoras" not in rule.fullname:
                LOGGER.warning(
                    "Trying to apply theorem when symbolic checks have failed."
                    f" Check the mappings of theorem '{rule}'."
                )
                continue
            valid_conclusions.add(conclusion_dep)
        return valid_conclusions

    def _match_generic(self, rule: Rule, proof: ProofState) -> set[Justification]:
        theorem_conclusions: set[Justification] = set()
        points = [p.name for p in proof.symbols.points]

        for mapping in self.theorem_mapper.mappings(rule, points, proof=proof):
            why: list[Predicate] = []
            assert len(mapping) > 0

            applicable = True
            for premise in rule.premises:
                premise_args = tuple(mapping[arg] for arg in premise.variables)
                premise_predicate_construction = (
                    PredicateConstruction.from_predicate_type_and_args(
                        PredicateType(premise.name), premise_args
                    )
                )
                premise_predicate = predicate_from_construction(
                    premise_predicate_construction, proof.symbols.points
                )
                if premise_predicate is None:
                    if "Pythagoras" not in rule.fullname:
                        LOGGER.warning(
                            f"Trying to build invalid premise '{premise_predicate}' of theorem '{rule}'."
                        )
                    applicable = False
                    break
                if not proof.check_numerical(premise_predicate):
                    if (
                        premise_predicate.predicate_type
                        in SYMBOLIC_PREDICATE_TYPES_NAMES
                    ):
                        LOGGER.warning(
                            f"Trying match numerically false premise '{premise_predicate}' of theorem '{rule}'."
                        )
                    applicable = False
                    break
                why.append(premise_predicate)
            if not applicable:
                continue

            for conclusion in rule.conclusions:
                conclusion_args = tuple(mapping[arg] for arg in conclusion.variables)
                conclusion_predicate_construction = (
                    PredicateConstruction.from_predicate_type_and_args(
                        PredicateType(conclusion.name), conclusion_args
                    )
                )
                conclusion_predicate = predicate_from_construction(
                    conclusion_predicate_construction,
                    points_registry=proof.symbols.points,
                )
                if conclusion_predicate is None:
                    LOGGER.debug(
                        f"Trying to build invalid conclusion '{conclusion_predicate}' of theorem '{rule}'."
                    )
                    continue
                theorem_conclusions.add(
                    RuleApplication(
                        predicate=conclusion_predicate, rule=rule, premises=tuple(why)
                    )
                )

        return theorem_conclusions


class TheoremMapper(ABC):
    @abstractmethod
    def mappings(
        self,
        rule: "Rule",
        points: list[PredicateArgument],
        proof: ProofState,
    ) -> Generator[Mapping, None, None]:
        """Find valid mappings for the given rule using the given points."""


class CCMapper(TheoremMapper):
    def mappings(
        self,
        rule: "Rule",
        points: list[PredicateArgument],
        proof: ProofState,
    ) -> Generator[Mapping, None, None]:
        variables = rule.variables
        for mapping in generate_permutations_as_dicts(points, variables):
            yield mapping


class FilterMapper(TheoremMapper):
    def mappings(
        self,
        rule: "Rule",
        points: list[PredicateArgument],
        proof: ProofState,
    ) -> Generator[Mapping, None, None]:
        # Each rule will have a collection of premises P.
        # S is the set of symbolic premises, and N is the set of
        # numerical premises; they partition P.
        # What we will do is iterate over all the filtered mappings for
        # the set of premises S. That will yield a set of variables compatible
        # with the premises S. For each of those, we will generate
        # a collection of mappings that also assign free points to
        # the variables in the set of premises N.

        LOGGER.debug("FilterMapper running for rule: %s", rule.fullname)

        for circle in proof.symbols.circles:
            LOGGER.debug("Circle: %s", circle)
            for point in circle.points:
                LOGGER.debug("  Point: %s", point)

        symbolic_premises: list[RuleConstruction] = []
        numerical_premises: list[RuleConstruction] = []

        variables_in_symbolic_premises: set[VarName] = set()
        for premise in rule.premises:
            premise_predicate = premise.name
            if premise_predicate not in SYMBOLIC_PREDICATE_TYPES_NAMES:
                numerical_premises.append(premise)
            else:
                symbolic_premises.append(premise)
                variables_in_symbolic_premises.update(premise.variables)

        numerical_only_variable_names_in_theorem: set[VarName] = (
            set(rule.variables) - variables_in_symbolic_premises
        )

        predicates_by_predicate_name = predictates_by_type_name_from_hypergraph(
            proof.graph.hyper_graph
        )

        LOGGER.debug("Symbolic premises: %s", predicates_by_predicate_name)
        LOGGER.debug("Numerical premises: %s", numerical_premises)

        if not rule.allow_point_repetition and len(rule.variables) > len(points):
            LOGGER.debug(
                "Skipping theorem %s because it has more variables than points",
                rule.fullname,
            )
            yield from []
            return

        symbolic_partial_mappings: list[PartialMapping] = []
        for premise in symbolic_premises:
            predicate_to_match = premise.name
            variables = premise.variables
            predicates_of_predicate: list[Predicate] = []
            match predicate_to_match:
                case "cyclic":
                    for circle in proof.symbols.circles:
                        if len(circle.points) < len(variables):
                            continue
                        for combo in itertools.combinations(
                            circle.points, len(variables)
                        ):
                            predicates_of_predicate.append(Cyclic(points=combo))
                case "coll":
                    for line in proof.symbols.lines:
                        if len(line.points) < len(variables):
                            continue
                        for combo in itertools.combinations(
                            line.points, len(variables)
                        ):
                            assert all(hasattr(e, "num") for e in combo)
                            predicates_of_predicate.append(Coll(points=combo))
                case _:
                    predicates_of_predicate = predicates_by_predicate_name.get(
                        PredicateType(predicate_to_match), []
                    )
            LOGGER.debug(
                "Have %s predicates for predicate %s",
                len(predicates_of_predicate),
                predicate_to_match,
            )

            if len(predicates_of_predicate) == 0:
                # If there is at least one predicate that has no predicates, then
                # there is no chance we can match this rule. So, we fail fast.
                yield from []
                return
            symbolic_partial_mappings.append(
                PartialMapping(
                    variables,
                    [efficient_version(_stmt) for _stmt in predicates_of_predicate],
                )
            )

        LOGGER.debug("Symbolic partial mappings: %s", symbolic_partial_mappings)

        LOGGER.debug(
            "Mapping %s variables against %s points",
            len(rule.variables),
            len(points),
        )
        LOGGER.debug("Premises: %s", rule.premises)

        empty_mapping: list[Mapping] = [{}]
        mapping_iterator: Iterator[Mapping] = (
            iterate_mappings(
                partial_mappings=symbolic_partial_mappings,
                remaining_variables=rule.variables,
                remaining_points=set(points),
                allow_point_repetition=rule.allow_point_repetition,
            )
            if symbolic_partial_mappings
            else iter(empty_mapping)
        )

        for mapping in mapping_iterator:
            if not numerical_only_variable_names_in_theorem:
                LOGGER.debug("Yielding mapping (195): %s", mapping)
                yield mapping
            else:
                if rule.allow_point_repetition:
                    available_points = set(points)
                else:
                    available_points = set(points) - set(mapping.values())
                for mapping_final in iterate_mapping_with_complementary_assignments(
                    mapping,
                    remaining_variables_in_theorem=list(
                        numerical_only_variable_names_in_theorem
                    ),
                    points=list(available_points),
                ):
                    LOGGER.debug(
                        "Yielding mapping with complementary assignments: %s %s",
                        rule.fullname,
                        mapping_final,
                    )
                    yield mapping_final


def predictates_by_type_name_from_hypergraph(
    hyper_graph: dict[Predicate, Justification],
) -> dict[PredicateType, list[Predicate]]:
    predicates_by_predictate_name: dict[PredicateType, list[Predicate]] = (
        collections.defaultdict(list)
    )
    for predicate in hyper_graph:
        predicates_by_predictate_name[predicate.predicate_type].append(predicate)
    return predicates_by_predictate_name


def iterate_mapping_with_complementary_assignments(
    mapping: Mapping, remaining_variables_in_theorem: list[str], points: list[str]
) -> Generator[Mapping, None, None]:
    """Yields mappings with additional variable assignments.

    For a given initial mapping of variables to points, yields either:
    1. Just the input mapping if there are no remaining unmapped variables
    2. New mappings that combine the input mapping with all possible assignments
    of the remaining unmapped variables to points

    Args:
        mapping: Initial dict mapping theorem variables to point names

    Yields:
        Iterator mapping all theorem variables to point names
    """
    if not remaining_variables_in_theorem:
        yield mapping
        return

    for mapping_of_remaining_variables in generate_permutations_as_dicts(  # type: ignore
        points, remaining_variables_in_theorem
    ):
        mapping.update(mapping_of_remaining_variables)  # type: ignore
        yield mapping


SYMBOLIC_PREDICATE_TYPES_NAMES = {
    symbolic_predicate_type.value
    for symbolic_predicate_type in SYMBOLIC_PREDICATES.keys()
}


LookupDict: TypeAlias = dict[VarName, dict[PredicateArgument, set[int]]]


class PartialMapping:
    lookup_dict: LookupDict  # pyright: ignore
    predicate_name: str
    predicates: list[EfficientStatement]
    vars_to_assign: tuple[VarName, ...]
    valid_arg_permutations: list[EfficientStatement]
    partial_assignment: dict[VarName, PredicateArgument]
    permutation_idxs_compatible_with_partial_assignment: set[int]
    mapping_hash: str

    def __init__(
        self,
        vars_to_assign: tuple[VarName, ...],
        predicates: list[EfficientStatement],
        lookup_dict: LookupDict | None = None,
        valid_arg_permutations: list[EfficientStatement] | None = None,
        partial_assignment: dict[VarName, PredicateArgument] | None = None,
        permutation_idxs_compatible_with_partial_assignment: set[int] | None = None,
    ):
        predicate_predicate_names: list[str] = [e[0] for e in predicates]
        assert len(set(predicate_predicate_names)) == 1, (
            f"PartialMapping must have predicates with the same predicate, but got {predicate_predicate_names}"
        )
        self.predicate_name = predicate_predicate_names[0]

        self.predicates = predicates
        self.vars_to_assign = vars_to_assign
        self._initialize_valid_arg_permutations(lookup_dict, valid_arg_permutations)

        self.partial_assignment = partial_assignment or dict()
        if permutation_idxs_compatible_with_partial_assignment is not None:
            self.permutation_idxs_compatible_with_partial_assignment = (
                permutation_idxs_compatible_with_partial_assignment
            )

    def with_assignment(
        self, var: VarName, point: PredicateArgument
    ) -> PartialMapping | None:
        # If the variable we're considering assigning is not even in our set of variables,
        # then it will be compatible with all of our assignments.
        if var not in self.vars_to_assign:
            return self

        # If we've already assigned the variable, that is an error, because we should only
        # be assigning new variables.
        if var in self.partial_assignment:
            raise ValueError(
                f"Attempting to re-assign variable {var} to {point} with {self.partial_assignment}"
            )

        updated_permutation_idxs_compatible_with_partial_assignment = (
            self.permutation_idxs_compatible_with_partial_assignment
            & self.lookup_dict[var][point]
        )
        if not updated_permutation_idxs_compatible_with_partial_assignment:
            return None
        updated_partial_assignment = {**self.partial_assignment, var: point}
        return PartialMapping(
            vars_to_assign=self.vars_to_assign,
            predicates=self.predicates,
            lookup_dict=self.lookup_dict,
            valid_arg_permutations=self.valid_arg_permutations,
            partial_assignment=updated_partial_assignment,
            permutation_idxs_compatible_with_partial_assignment=updated_permutation_idxs_compatible_with_partial_assignment,
        )

    def _initialize_valid_arg_permutations(
        self,
        lookup_dict: LookupDict | None,
        valid_arg_permutations: list[EfficientStatement] | None,
    ):
        assert not ((lookup_dict is None) ^ (valid_arg_permutations is None))
        if lookup_dict is not None:
            self.lookup_dict = lookup_dict
            self.valid_arg_permutations = (
                valid_arg_permutations if valid_arg_permutations is not None else []
            )
            return

        self.lookup_dict: LookupDict = collections.defaultdict(
            lambda: collections.defaultdict(set)
        )
        self.valid_arg_permutations = []

        points_iterator: Iterator[EfficientStatement]

        for predicate in self.predicates:
            points_iterator = generate_permutations(predicate)

            for stmt_permutation in points_iterator:
                assert len(self.vars_to_assign) == len(stmt_permutation) - 1, (
                    f"{self.predicate_name} {self.vars_to_assign} {stmt_permutation}"
                )
                permutation_pt_names = cast(
                    tuple[PredicateArgument, ...], stmt_permutation[1:]
                )
                self.valid_arg_permutations.append(stmt_permutation)
                if (
                    _get_valid_assignment_to_predicate_if_exists(
                        self.vars_to_assign, permutation_pt_names
                    )
                    is not None
                ):
                    for var, point in zip(self.vars_to_assign, permutation_pt_names):
                        self.lookup_dict[var][point].add(
                            len(self.valid_arg_permutations) - 1
                        )
        self.permutation_idxs_compatible_with_partial_assignment = set(
            range(len(self.valid_arg_permutations))
        )
        return self.valid_arg_permutations

    def __str__(self) -> str:
        return f"PartialMapping(predicate_name={self.predicate_name}, vars_to_assign={self.vars_to_assign})"

    def __repr__(self) -> str:
        return self.__str__()


def _get_valid_assignment_to_predicate_if_exists(
    vars: tuple[VarName, ...], points: tuple[PredicateArgument, ...]
) -> tuple[PredicateArgument, ...] | None:
    assignment: dict[VarName, PredicateArgument] = {}
    for v, p in zip(vars, points):
        if v in assignment and assignment[v] != p:
            return None
        assignment[v] = p
    return tuple(assignment[v] for v in vars)


def iterate_mappings(
    partial_mappings: list[PartialMapping],
    remaining_variables: list[VarName],
    remaining_points: set[PredicateArgument],
    current_assignment: dict[VarName, PredicateArgument] | None = None,
    conclusions_yielded: set[EfficientStatement] | None = None,
    allow_point_repetition: bool = False,
) -> Iterator[dict[VarName, PredicateArgument]]:
    current_assignment = current_assignment or {}
    if not remaining_variables:
        yield current_assignment
        return

    conclusions_yielded = conclusions_yielded or set()

    var, *remaining_vars_next_level = remaining_variables

    prefix = "  " * len(current_assignment)
    LOGGER.debug("%sChecking variable %s", prefix, var)
    for pt in remaining_points:
        LOGGER.debug("%sChecking point %s", prefix, pt)

        next_partial_mappings: list[PartialMapping] = []
        for current_partial_mapping in partial_mappings:
            next_partial_mapping = current_partial_mapping.with_assignment(var, pt)
            if next_partial_mapping is None:
                # If it's none, it means there was no assignment of current_assignment + var = pt
                # that was compatible with the partial mapping, so we must backtrack from
                # this partial assignment.
                break
            else:
                next_partial_mappings.append(next_partial_mapping)

        if len(next_partial_mappings) < len(partial_mappings):
            # This means that at least one partial mapping was unsatisfied, so we go on to the next point at this level.
            continue

        # If we've gotten here, it means that {**current_assignment, var: pt} is a valid assignment
        # for all the partial mappings, so we can continue to assign the rest of the variables, using
        # this prefix.

        if allow_point_repetition:
            new_remaining_points = remaining_points
        else:
            new_remaining_points = remaining_points - {pt}

        for mapping in iterate_mappings(
            partial_mappings=next_partial_mappings,
            remaining_variables=remaining_vars_next_level,
            remaining_points=new_remaining_points,
            current_assignment={**current_assignment, var: pt},
            allow_point_repetition=allow_point_repetition,
        ):
            yield mapping



from __future__ import annotations

from typing import Literal, Optional, TypeVar

from newclid.numerical import close_enough
from newclid.numerical.check import same_clock
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Triangle

T = TypeVar("T", bound=str)


def two_triangles(
    a: T, b: T, c: T, p: T, q: T, r: T
) -> Optional[tuple[T, T, T, T, T, T]]:
    if a == b or a == c or b == c or p == q or p == r or q == r:
        return None
    if a == p and b == q and c == r:
        return None

    (a0, p0), (b0, q0), (c0, r0) = sorted(((a, p), (b, q), (c, r)))
    (a1, p1), (b1, q1), (c1, r1) = sorted(((p, a), (q, b), (r, c)))
    return min((a0, b0, c0, p0, q0, r0), (a1, b1, c1, p1, q1, r1))


class SimtriClock(PredicateInterface):
    """simtri A B C P Q R -

    Represent that triangles ABC and PQR are similar under orientation-preserving
    transformations taking A to P, B to Q and C to R.

    It is equivalent to the three eqangle and eqratio predicates
    on the corresponding angles and sides.
    """

    predicate_type: Literal[PredicateType.SIMTRI_CLOCK] = PredicateType.SIMTRI_CLOCK
    triangle1: Triangle
    triangle2: Triangle

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> Optional[tuple[PredicateArgument, ...]]:
        return two_triangles(*args)

    def check_numerical(self) -> bool:
        a, b, c = self.triangle1
        p, q, r = self.triangle2

        ab = a.num.distance(b.num)
        ac = a.num.distance(c.num)
        bc = b.num.distance(c.num)
        pq = p.num.distance(q.num)
        pr = p.num.distance(r.num)
        qr = q.num.distance(r.num)
        return (
            close_enough(ab * pr, ac * pq)
            and close_enough(bc * pr, ac * qr)
            and same_clock(a.num, b.num, c.num, p.num, q.num, r.num)
        )

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.triangle1 + self.triangle2)

    def __str__(self) -> str:
        a, b, c = self.triangle1
        p, q, r = self.triangle2
        return f"▲{a}{b}{c} ≅ ▲{p}{q}{r}"


class SimtriReflect(PredicateInterface):
    """simtrir A B C P Q R -

    Represent that triangles ABC and PQR are similar under orientation-preserving
    transformations taking A to P, B to Q and C to R.

    It is equivalent to the three eqangle and eqratio predicates
    on the corresponding angles and sides.
    """

    predicate_type: Literal[PredicateType.SIMTRI_REFLECT] = PredicateType.SIMTRI_REFLECT
    triangle1: Triangle
    triangle2: Triangle

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return two_triangles(*args)

    def check_numerical(self) -> bool:
        a, b, c = self.triangle1
        p, q, r = self.triangle2
        ab = a.num.distance(b.num)
        ac = a.num.distance(c.num)
        bc = b.num.distance(c.num)
        pq = p.num.distance(q.num)
        pr = p.num.distance(r.num)
        qr = q.num.distance(r.num)
        return (
            close_enough(ab * pr, ac * pq)
            and close_enough(bc * pr, ac * qr)
            and same_clock(a.num, b.num, c.num, p.num, r.num, q.num)
        )

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.triangle1 + self.triangle2)

    def __str__(self) -> str:
        a, b, c = self.triangle1
        p, q, r = self.triangle2
        return f"▲{a}{b}{c} ≅ ▲{p}{q}{r}"



import logging
from typing import Literal

import numpy as np
from pydantic import BaseModel

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.heuristics.geometric_objects import LineHeuristic
from newclid.jgex.clause import JGEXClause
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType

LOGGER = logging.getLogger(__name__)


class ThreeCircumcirclesForEqangleGoalHeuristicConfig(BaseModel):
    heuristic_name: Literal[HeuristicName.THREE_CIRCUMCIRCLES_FOR_EQANGLE_GOAL] = (
        HeuristicName.THREE_CIRCUMCIRCLES_FOR_EQANGLE_GOAL
    )


class ThreeCircumcirclesForEqangleGoalHeuristic(Heuristic):
    """
    Given an eqangle goal, create the vertices and implement the two versions of the 3 circumcircles method for the corresponding quadrilateral configuration.

    Check the solution of "Langley's Adventious Angles" problem for more details.
    """

    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: np.random.Generator,
    ) -> list[JGEXClause]:
        i = 0
        eqangle_goals = [
            g for g in setup.goals if g.predicate_type == PredicateType.EQUAL_ANGLES
        ]
        added_points: list[PredicateArgument] = []
        new_clauses: list[JGEXClause] = []

        while i < max_new_points and eqangle_goals:
            goal_index = rng.choice(len(eqangle_goals))
            goal = eqangle_goals.pop(goal_index)

            line_matrix: list[list[LineHeuristic]] = [
                [LineHeuristic(points=()), LineHeuristic(points=())],
                [LineHeuristic(points=()), LineHeuristic(points=())],
            ]

            for line in setup.lines:
                if goal.args[0] in line.points and goal.args[1] in line.points:
                    line_matrix[0][0] = line
                if goal.args[2] in line.points and goal.args[3] in line.points:
                    line_matrix[0][1] = line
                if goal.args[4] in line.points and goal.args[5] in line.points:
                    line_matrix[1][0] = line
                if goal.args[6] in line.points and goal.args[7] in line.points:
                    line_matrix[1][1] = line

            # TODO: Clarify what this does
            vertices: list[PredicateArgument] = []
            for j in [0, 2]:
                for k in [4, 6]:
                    line_0 = line_matrix[0][int(j % 2)]
                    line_1 = line_matrix[1][int((k / 2) % 2)]
                    if set(line_0.points) & set(line_1.points):
                        existing_vertex = next(
                            iter(set(line_0.points) & set(line_1.points))
                        )
                        vertices.append(existing_vertex)
                    else:
                        vertex = get_available_from_alphabet(
                            list(setup.points) + added_points
                        )
                        if vertex is None:
                            LOGGER.warning(
                                "No available points left in the alphabet, skipping vertex addition."
                            )
                            continue
                        added_points.append(vertex)

                        new_clause_string = f"{vertex} = on_line {vertex} {goal.args[j]} {goal.args[j + 1]}, on_line {vertex} {goal.args[k]} {goal.args[k + 1]}"
                        new_clause = JGEXClause.from_str(new_clause_string)[0]
                        new_clauses.append(new_clause)

            i += 4

            if len(vertices) < 4:
                LOGGER.warning(
                    "Not enough vertices created for the full construction. We stop here."
                )
                continue

            centers: list[PredicateArgument] = []
            center_1 = get_available_from_alphabet(list(setup.points) + added_points)
            if center_1 is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping center addition."
                )
            else:
                added_points.append(center_1)
                new_clause_string = f"{center_1} = circle {center_1} {vertices[1]} {vertices[2]} {vertices[0]}"
                new_clause = JGEXClause.from_str(new_clause_string)[0]
                new_clauses.append(new_clause)

            center_2 = get_available_from_alphabet(list(setup.points) + added_points)
            if center_2 is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping center addition."
                )
            else:
                added_points.append(center_2)
                new_clause_string = f"{center_2} = circle {center_2} {vertices[1]} {vertices[2]} {vertices[3]}"
                new_clause = JGEXClause.from_str(new_clause_string)[0]
                new_clauses.append(new_clause)

            i += 2
            if len(centers) < 2:
                LOGGER.warning(
                    "Not enough centers created for the full construction. We stop here."
                )
                continue

            center_3 = get_available_from_alphabet(list(setup.points) + added_points)
            if center_3 is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping center addition."
                )
            else:
                added_points.append(center_3)
                new_clause_string = f"{center_3} = circle {center_3} {center_1} {center_2} {vertices[1]}"
                new_clause = JGEXClause.from_str(new_clause_string)[0]
                new_clauses.append(new_clause)

            center_4 = get_available_from_alphabet(list(setup.points) + added_points)
            if center_4 is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping center addition."
                )
            else:
                added_points.append(center_4)
                new_clause_string = f"{center_4} = circle {center_4} {center_1} {center_2} {vertices[2]}"
                new_clause = JGEXClause.from_str(new_clause_string)[0]
                new_clauses.append(new_clause)

            i += 2

        return new_clauses



"""Command line interface for Newclid."""

from __future__ import annotations

import logging
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    Namespace,
    _SubParsersAction,  # pyright: ignore
)
from enum import Enum
from pathlib import Path
from typing import Any

import numpy as np
from pydantic import BaseModel

from newclid.agent._index import AgentName
from newclid.api import ProblemBuilder
from newclid.ggb.problem_builder import GeogebraProblemBuilder
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.problem import PredicateConstruction


class NewclidOptions(BaseModel):
    output_dir: Path | None
    saturate: bool
    agent: AgentName | None
    seed: int | None
    log_level: LogLevel


def parse_cli_args(args: list[str]) -> tuple[NewclidOptions, ProblemBuilder]:
    parser = make_cli_parser()
    parsed_args = parser.parse_args(args)
    problem_builder = parsed_args.make_problem_builder(parsed_args)
    args_dict = vars(parsed_args)
    _adapt_log_level(args_dict)
    return NewclidOptions.model_validate(args_dict), problem_builder


def make_cli_parser() -> ArgumentParser:
    parser = ArgumentParser("newclid", formatter_class=ArgumentDefaultsHelpFormatter)
    _add_top_parser_arguments(parser)
    subparsers = parser.add_subparsers(
        title="Problem builders",
        description="Different methods to build a newclid problem.",
    )
    _add_jgex_parser(subparsers)
    _add_ggb_parser(subparsers)
    return parser


class LogLevel(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR


def _adapt_log_level(args_dict: dict[str, Any]) -> None:
    if "log_level" in args_dict:
        args_dict["log_level"] = LogLevel[args_dict["log_level"]]


def _add_top_parser_arguments(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument("--output-dir", "-o", default=None, help="Output directory")
    parser.add_argument(
        "--saturate",
        action="store_true",
        help="Ignore the goal and saturate the problem",
    )
    parser.add_argument(
        "--draw-initial-figure",
        action="store_true",
        help="Draw the initial figure",
    )
    parser.add_argument(
        "--draw-final-figure",
        action="store_true",
        help="Draw the final figure",
    )
    parser.add_argument(
        "--agent",
        default=None,
        help="Name of the agent to use",
        choices=[name.value for name in AgentName],
    )
    parser.add_argument(
        "--seed", default=None, type=int, help="Seed for random sampling"
    )
    parser.add_argument(
        "--log-level",
        default=LogLevel.INFO.name,
        type=str,
        choices=[level.name for level in LogLevel],
        help="Logging level c.f. https://docs.python.org/3/library/logging.html#logging-levels",
    )
    return parser


def _jgex_problem_builder(args: Namespace) -> ProblemBuilder:
    if args.problem_id is None and args.problem is None:  # pragma: no cover
        raise ValueError("Either --problem-id or --problem must be provided")
    if args.problem_id is not None and args.problem is not None:  # pragma: no cover
        raise ValueError("Only one of --problem-id or --problem must be provided")
    jgex_problem_builder = JGEXProblemBuilder(rng=np.random.default_rng(args.seed))
    if args.problem_id is not None:
        jgex_problem_builder.with_problem_from_file(
            problem_name=args.problem_id, problems_path=Path(args.file)
        )
    else:
        jgex_problem_builder.with_problem_from_txt(args.problem)
    return jgex_problem_builder


def _add_jgex_parser(subparsers: _SubParsersAction[ArgumentParser]) -> ArgumentParser:
    parser = subparsers.add_parser("jgex")
    parser.add_argument("--problem-id", "-p", help="The id of a JGEX problem to solve")
    parser.add_argument(
        "--file",
        "-f",
        help="The name of the problems file to read the problem from",
    )
    parser.add_argument("--problem", "-s", help="A JGEX string of a problem.")
    parser.set_defaults(make_problem_builder=_jgex_problem_builder)
    return parser


def _ggb_problem_builder(args: Namespace) -> ProblemBuilder:
    return GeogebraProblemBuilder(ggb_file_path=Path(args.file)).with_goals(
        [PredicateConstruction.from_str(g) for g in args.goals]
    )


def _add_ggb_parser(subparsers: _SubParsersAction[ArgumentParser]) -> ArgumentParser:
    parser = subparsers.add_parser("ggb")
    parser.add_argument(
        "--file",
        "-f",
        help="The ggb export file to read the problem from",
        required=True,
    )
    parser.add_argument(
        "--goals",
        "-g",
        nargs="+",
        help="The goals to add to the problem",
        type=str,
        required=True,
    )
    parser.set_defaults(make_problem_builder=_ggb_problem_builder)
    return parser



"""Module to compute the justifications of the goals."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from newclid.justifications._index import JustificationError
from newclid.justifications.justification import justify_dependency

if TYPE_CHECKING:
    from newclid.justifications.justification import Justification
    from newclid.predicates import Predicate
    from newclid.proof_state import ProofState

LOGGER = logging.getLogger(__name__)


def goals_justifications(
    goals: list[Predicate], proof_state: ProofState
) -> tuple[tuple[Justification, ...], dict[Predicate, bool]]:
    sub_proof: dict[Predicate, tuple[Justification, ...]] = {}
    proof_steps: list[Justification] = []
    goal_is_proven: dict[Predicate, bool] = {}
    for goal in goals:
        # We don't need to re-calculate if the goal's subproof is already memoized.
        if goal in sub_proof:
            goal_is_proven[goal] = True
            continue
        try:
            proof_of_goal = _proof_of_predicate(goal, proof_state, sub_proof)
            goal_is_proven[goal] = True
        except JustificationError:
            # Could not justify this goal, so we skip it
            goal_is_proven[goal] = False
            continue
        for proof_step in proof_of_goal:
            if proof_step not in proof_steps:
                proof_steps.append(proof_step)
    return tuple(proof_steps), goal_is_proven


def _proof_of_predicate(
    predicate: Predicate,
    proof_state: ProofState,
    sub_proof: dict[Predicate, tuple[Justification, ...]],
) -> tuple[Justification, ...]:
    stack: list[Predicate] = [predicate]
    while stack:
        current_predicate = stack[-1]  # Peek
        if current_predicate in sub_proof:
            stack.pop()
            continue

        justification = _get_justification_of_predicate(current_predicate, proof_state)
        premises = justify_dependency(justification, proof_state)
        unresolved_premises = [p for p in premises if p not in sub_proof]

        if not unresolved_premises:
            stack.pop()
            cur_proof: tuple[Justification, ...] = ()
            for p in premises:
                # Reconstruct the full subproof for the premise
                p_dep = _get_justification_of_predicate(p, proof_state)
                premise_full_proof = sub_proof[p] + (p_dep,)
                cur_proof += tuple(
                    dep for dep in premise_full_proof if dep not in cur_proof
                )
            sub_proof[current_predicate] = cur_proof
        else:
            stack.extend(unresolved_premises)

    return sub_proof[predicate] + (proof_state.graph.hyper_graph[predicate],)


def _get_justification_of_predicate(
    predicate: Predicate, proof_state: ProofState
) -> Justification:
    justification = proof_state.graph.hyper_graph.get(predicate)
    if justification is None:
        LOGGER.debug(f"Justification not found for {predicate}.")
        justification = proof_state.justify(predicate)
    return justification



import logging
from typing import Literal

import numpy
from pydantic import BaseModel

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.jgex.clause import JGEXClause

LOGGER = logging.getLogger(__name__)


class ReflectOnCenterHeuristicConfig(BaseModel):
    heuristic_name: Literal[HeuristicName.REFLECT_ON_CENTER] = (
        HeuristicName.REFLECT_ON_CENTER
    )


class ReflectOnCenterHeuristic(Heuristic):
    """Reflect points on the center of a circle."""

    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: numpy.random.Generator,
    ) -> list[JGEXClause]:
        circles = list(setup.circles)
        points = list(setup.points)
        new_clauses: list[JGEXClause] = []
        for _ in range(max_new_points):
            if not circles:
                break

            circle_index = rng.choice(len(circles))
            circle = circles.pop(circle_index)

            p1 = circle.points[0]
            p2 = circle.points[1]
            p3 = circle.points[2]

            new_center = get_available_from_alphabet(points)
            if new_center is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping reflection on this center."
                )
                continue

            points.append(new_center)

            new_clause_string = f"{new_center} = foot {new_center} {p1} {p2} {p3}"
            new_clause = JGEXClause.from_str(new_clause_string)[0]
            new_clauses.append(new_clause)

            # Reflect all points on the center of the circle
            for point in circle.points:
                new_point = get_available_from_alphabet(points)
                if new_point is None:
                    LOGGER.warning(
                        "No available points left in the alphabet, skipping reflection addition."
                    )
                    continue
                points.append(new_point)

                new_clause_string = (
                    f"{new_point} = mirror {new_point} {point} {new_center}"
                )
                new_clause = JGEXClause.from_str(new_clause_string)[0]
                new_clauses.append(new_clause)

        return new_clauses



import logging

import numpy as np
from newclid.api import GeometricSolverBuilder
from newclid.jgex.problem_builder import JGEXProblemBuilder

LOGGER = logging.getLogger(__name__)


def test_can_write_partial_proof():
    problem_txt = (
        "a b c = triangle a b c"
        "; d = angle_bisector d b a c"
        "; e = on_aline e d a d c b, on_line e a c"
        "; f = on_aline f d a d b c, on_line f a b"
        "; x = on_bline x b c, on_line x a c"
        "; o1 = circle o1 a d c"
        "; o2 = circle o2 e x d"
        "; y = on_line y e f, on_line y b c"
        " ? eqangle c x b c b c b x"
        "; coll o1 o2 y"
    )
    rng = np.random.default_rng(42)
    problem_builder = JGEXProblemBuilder(rng=rng).with_problem_from_txt(problem_txt)
    solver = GeometricSolverBuilder(rng=rng).build(problem_builder.build())
    solver.run()
    assert solver.run_infos is not None
    assert solver.run_infos.success_per_goal["∠(BC,BX) = ∠(CX,BC) succeeded"]
    assert not solver.run_infos.success_per_goal["O₁, O₂, Y are collinear succeeded"]
    partial_proof = solver.proof()
    LOGGER.debug(partial_proof)
    assert "∠(BC,BX) = ∠(CX,BC) : Proved [C0]" in partial_proof
    assert "O₁, O₂, Y are collinear : Could not prove" in partial_proof


def test_circle_merge_is_written():
    problem_txt = (
        "a b c = triangle a b c"
        "; m = midpoint m a b"
        "; n = midpoint n b c"
        "; p = midpoint p a c"
        "; f1 = foot f1 a b c"
        "; f2 = foot f2 b a c"
        "; f3 = foot f3 c a b"
        " ? cyclic m n p f1 f2 f3"
    )
    rng = np.random.default_rng(42)
    problem_builder = JGEXProblemBuilder(rng=rng).with_problem_from_txt(problem_txt)
    solver = GeometricSolverBuilder(rng=rng).build(problem_builder.build())
    success = solver.run()
    assert success
    assert solver.run_infos is not None
    partial_proof = solver.proof()
    LOGGER.debug(partial_proof)
    assert "Circle merge" in partial_proof



import pytest
from newclid.justifications._index import JustificationError
from newclid.justifications.justification import DirectConsequence, Justification
from newclid.numerical.geometries import PointNum
from newclid.predicate_types import PredicateArgument
from newclid.predicates import Predicate
from newclid.predicates.collinearity import Coll
from newclid.problem import ProblemSetup
from newclid.proof_justifications import _proof_of_predicate  # type: ignore
from newclid.proof_state import ProofState
from newclid.symbols.points_registry import Point
from pytest_mock import MockerFixture


def p(name: str) -> PredicateArgument:
    return PredicateArgument(name)


class TestProofDependencies:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.proof_state = ProofState(
            problem=ProblemSetup(
                points=(
                    Point(name=p("a"), num=PointNum(x=0.0, y=0.0)),
                    Point(name=p("b"), num=PointNum(x=1.0, y=0.0)),
                    Point(name=p("c"), num=PointNum(x=2.0, y=0.0)),
                    Point(name=p("d"), num=PointNum(x=3.0, y=0.0)),
                    Point(name=p("e"), num=PointNum(x=4.0, y=0.0)),
                    Point(name=p("f"), num=PointNum(x=5.0, y=0.0)),
                ),
                assumptions=(),
                goals=(),
            ),
            rule_matcher=None,  # type: ignore
            deductors=[],
            rng=42,
        )
        self.pred_graph = self.proof_state.graph

        a, b, c, d, e, f = self.proof_state.symbols.points.names2points(
            [PredicateArgument(name) for name in ("a", "b", "c", "d", "e", "f")]
        )
        self.pred_A = Coll(points=(a, b, c))
        self.pred_B = Coll(points=(b, c, d))
        self.pred_C = Coll(points=(c, d, e))
        self.pred_D = Coll(points=(d, e, f))

    def test_proof_of_predicate_linear_proof(self):
        """Test a simple linear proof structure: A -> B -> C."""
        dep_A = DirectConsequence(predicate=self.pred_A, premises=())
        dep_B_from_A = DirectConsequence(predicate=self.pred_B, premises=(self.pred_A,))
        dep_C_from_B = DirectConsequence(predicate=self.pred_C, premises=(self.pred_B,))

        self.pred_graph.hyper_graph = {
            self.pred_A: dep_A,
            self.pred_B: dep_B_from_A,
            self.pred_C: dep_C_from_B,
        }

        sub_proof: dict[Predicate, tuple[Justification, ...]] = {}
        proof = _proof_of_predicate(self.pred_C, self.proof_state, sub_proof=sub_proof)

        assert proof == (dep_A, dep_B_from_A, dep_C_from_B)
        assert sub_proof[self.pred_A] == ()
        assert sub_proof[self.pred_B] == (dep_A,)
        assert sub_proof[self.pred_C] == (dep_A, dep_B_from_A)

    def test_proof_of_predicate_branching_proof(self):
        """Test a branching proof structure: A, B -> C."""
        # B is a premise here
        dep_A = DirectConsequence(predicate=self.pred_A, premises=())
        dep_B = DirectConsequence(predicate=self.pred_B, premises=())
        dep_C_from_AB = DirectConsequence(
            predicate=self.pred_C, premises=(self.pred_A, self.pred_B)
        )

        self.pred_graph.hyper_graph = {
            self.pred_A: dep_A,
            self.pred_B: dep_B,
            self.pred_C: dep_C_from_AB,
        }

        sub_proof: dict[Predicate, tuple[Justification, ...]] = {}
        proof = _proof_of_predicate(self.pred_C, self.proof_state, sub_proof=sub_proof)

        # The order of premises A and B in the proof depends on iteration order.
        expected_proof = (dep_A, dep_B, dep_C_from_AB)
        assert set(proof) == set(expected_proof)
        assert len(proof) == len(expected_proof)

        assert sub_proof[self.pred_A] == ()
        assert sub_proof[self.pred_B] == ()
        assert set(sub_proof[self.pred_C]) == set((dep_A, dep_B))

    def test_proof_of_predicate_dag_proof(self):
        """Test a DAG structure: A -> B, A -> C, (B, C) -> D."""
        dep_A = DirectConsequence(predicate=self.pred_A, premises=())
        dep_B_from_A = DirectConsequence(predicate=self.pred_B, premises=(self.pred_A,))
        dep_C_from_A = DirectConsequence(predicate=self.pred_C, premises=(self.pred_A,))
        dep_D_from_BC = DirectConsequence(
            predicate=self.pred_D, premises=(self.pred_B, self.pred_C)
        )

        self.pred_graph.hyper_graph = {
            self.pred_A: dep_A,
            self.pred_B: dep_B_from_A,
            self.pred_C: dep_C_from_A,
            self.pred_D: dep_D_from_BC,
        }

        sub_proof: dict[Predicate, tuple[Justification, ...]] = {}
        proof = _proof_of_predicate(self.pred_D, self.proof_state, sub_proof=sub_proof)

        expected_proof_without_duplicates = (
            dep_A,
            dep_B_from_A,
            dep_C_from_A,
            dep_D_from_BC,
        )
        assert proof == expected_proof_without_duplicates

        assert sub_proof[self.pred_A] == ()
        assert sub_proof[self.pred_B] == (dep_A,)
        assert sub_proof[self.pred_C] == (dep_A,)
        assert sub_proof[self.pred_D] == (dep_A, dep_B_from_A, dep_C_from_A)

    def test_dependency_not_found_in_hyper_graph(self, mocker: MockerFixture):
        """Test re-evaluating dependency with predicate.why()."""
        dep_A = DirectConsequence(predicate=self.pred_A, premises=())

        self.pred_graph.hyper_graph = {
            self.pred_A: dep_A,
            # dep_B_from_A is missing
        }

        sub_proof: dict[Predicate, tuple[Justification, ...]] = {}
        mocker.patch.object(
            self.proof_state.symbols.lines,
            "why_colllinear",
            return_value=DirectConsequence(
                predicate=self.pred_B, premises=(self.pred_A,)
            ),
        )
        proof = _proof_of_predicate(self.pred_B, self.proof_state, sub_proof=sub_proof)

        dep_B_from_A = DirectConsequence(predicate=self.pred_B, premises=(self.pred_A,))
        assert proof == (dep_A, dep_B_from_A)
        assert self.pred_B in self.pred_graph.hyper_graph
        assert self.pred_graph.hyper_graph[self.pred_B] == dep_B_from_A

    def test_dependency_not_found_and_why_fails(self):
        """Test case where dependency is missing and predicate.why() fails."""
        self.pred_graph.hyper_graph = {}
        sub_proof: dict[Predicate, tuple[Justification, ...]] = {}
        with pytest.raises(JustificationError):
            _proof_of_predicate(self.pred_A, self.proof_state, sub_proof=sub_proof)


class DummyPredicate:
    """A dummy predicate for testing purposes."""

    def __init__(self, name: str) -> None:
        self.name = name
        self.preloaded_reason: str | None = None
        self.preloaded_why: tuple[Predicate, ...] | None = None
        self.why_called_count: int = 0

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return args

    def add(self, proof_state: ProofState) -> tuple[Predicate, ...]:
        return ()

    def check(self, proof_state: ProofState) -> bool:
        return True

    def check_numerical(self) -> bool:
        return True

    def __str__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return hash(self.name)



from typing import Iterable

from matplotlib.artist import Artist
from matplotlib.axes import Axes

from newclid.draw.geometries import draw_circle, draw_triangle
from newclid.draw.predicates import draw_line, draw_line_symbol, draw_perp_rectangle
from newclid.draw.theme import DrawTheme
from newclid.jgex.clause import JGEXConstruction
from newclid.jgex.formulation import JGEXFormulation
from newclid.symbols.symbols_registry import SymbolsRegistry


def draw_jgex_problem_clauses(
    ax: Axes,
    jgex_problem: JGEXFormulation,
    symbols_registry: SymbolsRegistry,
    theme: DrawTheme,
):
    for clause in jgex_problem.clauses:
        for construction in clause.constructions:
            draw_jgex_constructions(
                ax, construction, symbols_registry=symbols_registry, theme=theme
            )


def draw_jgex_constructions(
    ax: Axes,
    construction: JGEXConstruction,
    symbols_registry: SymbolsRegistry,
    theme: DrawTheme,
) -> Iterable[Artist]:
    match construction.name:
        case "circle":
            x, a, _b, _c = symbols_registry.points.names2points(construction.args)  # type: ignore
            return [
                draw_circle(
                    ax=ax,
                    center=(x.num.x, x.num.y),
                    radius=x.num.distance(a.num),
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                )
            ]
        case "on_circle":
            _x, o, a = symbols_registry.points.names2points(construction.args)  # type: ignore
            radius = o.num.distance(a.num)
            return [
                draw_circle(
                    ax=ax,
                    center=(o.num.x, o.num.y),
                    radius=radius,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                )
            ]
        case "triangle":
            a, b, c = symbols_registry.points.names2points(construction.args)  # type: ignore
            return [
                draw_triangle(
                    ax=ax,
                    p0=a.num,
                    p1=b.num,
                    p2=c.num,
                    line_color=theme.triangle_color,
                    line_width=theme.thick_line_width,
                )
            ]
        case "on_line":
            _x, a, b = symbols_registry.points.names2points(construction.args)  # type: ignore
            return [
                draw_line(
                    ax=ax,
                    p0=a.num,
                    p1=b.num,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                )
            ]
        case "on_tline":
            x, y, a, b = symbols_registry.points.names2points(construction.args)  # type: ignore
            xy = symbols_registry.lines.line_thru_pair(x, y)
            ab = symbols_registry.lines.line_thru_pair(a, b)
            return [
                draw_line_symbol(
                    ax=ax,
                    line=xy,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=ab,
                    line_color=theme.line_color,
                    line_width=theme.thin_line_width,
                    ls=":",
                ),
                draw_perp_rectangle(
                    ax=ax,
                    line0=xy,
                    line1=ab,
                    color=theme.perpendicular_color,
                ),
            ]
        case "on_pline":
            x, y, a, b = symbols_registry.points.names2points(construction.args)  # type: ignore
            return [
                draw_line(
                    ax=ax,
                    p0=a.num,
                    p1=b.num,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line(
                    ax=ax,
                    p0=x.num,
                    p1=y.num,
                    line_color=theme.line_color,
                    line_width=theme.thin_line_width,
                    ls=":",
                ),
            ]
        case "on_dia":
            x, a, b = symbols_registry.points.names2points(construction.args)  # type: ignore
            xa = symbols_registry.lines.line_thru_pair(x, a)
            xb = symbols_registry.lines.line_thru_pair(x, b)
            return [
                draw_line_symbol(
                    ax=ax,
                    line=xa,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=xb,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_perp_rectangle(
                    ax=ax,
                    line0=xa,
                    line1=xb,
                    color=theme.perpendicular_color,
                ),
            ]
        case "midpoint":
            x, a, b = symbols_registry.points.names2points(construction.args)  # type: ignore
            ab = symbols_registry.lines.line_thru_pair(a, b)
            return [
                draw_line_symbol(
                    ax=ax,
                    line=ab,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
            ]
        case "segment":
            a, b = symbols_registry.points.names2points(construction.args)  # type: ignore
            return [
                draw_line(
                    ax=ax,
                    p0=a.num,
                    p1=b.num,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                )
            ]
        case "foot":
            f, a, b, c = symbols_registry.points.names2points(construction.args)  # type: ignore
            af = symbols_registry.lines.line_thru_pair(a, f)
            bc = symbols_registry.lines.line_thru_pair(b, c)
            return [
                draw_line_symbol(
                    ax=ax,
                    line=af,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=bc,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_perp_rectangle(
                    ax=ax, line0=af, line1=bc, color=theme.perpendicular_color
                ),
            ]
        case "incenter":
            d, a, b, c = symbols_registry.points.names2points(construction.args)  # type: ignore
            ad = symbols_registry.lines.line_thru_pair(a, d)
            bd = symbols_registry.lines.line_thru_pair(b, d)
            cd = symbols_registry.lines.line_thru_pair(c, d)
            return [
                draw_triangle(
                    ax,
                    a.num,
                    b.num,
                    c.num,
                    line_color=theme.triangle_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=ad,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=bd,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=cd,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
            ]
        case "excenter":
            d, a, b, c = symbols_registry.points.names2points(construction.args)  # type: ignore
            ad = symbols_registry.lines.line_thru_pair(a, d)
            bd = symbols_registry.lines.line_thru_pair(b, d)
            cd = symbols_registry.lines.line_thru_pair(c, d)
            return [
                draw_triangle(
                    ax,
                    a.num,
                    b.num,
                    c.num,
                    line_color=theme.triangle_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=ad,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=bd,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
                draw_line_symbol(
                    ax=ax,
                    line=cd,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                ),
            ]
        case _:
            return []



"""
Efficient statement handling for geometric theorem proving.

This module provides data structures and utilities for representing and manipulating
geometric statements in a theorem proving system. It includes:

- Named tuple classes for different types of geometric statements (perpendicular,
  parallel, equal angles, etc.)
- Named tuple classes for corresponding predicates with variable names
- Functions for generating permutations of statements based on geometric symmetries
- Functions for normalizing statements to canonical forms
- Utilities for finding representatives of equivalence classes

The module is designed for efficiency in geometric theorem proving by providing
canonical representations and symmetry handling for geometric statements.
"""

from __future__ import annotations

import itertools
import typing
from typing import Iterator, Literal, TypeAlias, Union

from newclid.predicate_types import PredicateArgument
from newclid.predicates import Predicate
from newclid.predicates._index import PredicateType
from newclid.rule import VarName


def efficient_version(predicate: Predicate) -> EfficientStatement:
    match predicate.predicate_type:
        case PredicateType.EQUAL_ANGLES:
            return EqangleStatement("eqangle", *predicate.to_tokens())
        case PredicateType.PERPENDICULAR:
            return PerpStatement("perp", *predicate.to_tokens())
        case PredicateType.PARALLEL:
            return ParaStatement("para", *predicate.to_tokens())
        case PredicateType.CONGRUENT:
            return CongStatement("cong", *predicate.to_tokens())
        case PredicateType.MIDPOINT:
            return MidptStatement("midp", *predicate.to_tokens())
        case PredicateType.EQUAL_RATIOS:
            return EqratioStatement("eqratio", *predicate.to_tokens())
        case PredicateType.COLLINEAR:
            return CollStatement("coll", *predicate.to_tokens())
        case PredicateType.CIRCUMCENTER:
            return CircleStatement("circle", *predicate.to_tokens())
        case PredicateType.CYCLIC:
            return CyclicStatement("cyclic", *predicate.to_tokens())
        case PredicateType.SIMTRI_CLOCK | PredicateType.SIMTRI_REFLECT:
            return TriangleStatement(
                predicate.predicate_type.value, *predicate.to_tokens()
            )
        case _:
            raise NotImplementedError(f"Unknown predicate: {predicate.predicate_type}")


class PerpStatement(typing.NamedTuple):
    name: Literal["perp"]
    a: PredicateArgument
    b: PredicateArgument
    c: PredicateArgument
    d: PredicateArgument


class PerpPredicate(typing.NamedTuple):
    name: Literal["perp"]
    A: VarName
    B: VarName
    C: VarName
    D: VarName


class ParaStatement(typing.NamedTuple):
    name: Literal["para"]
    a: PredicateArgument
    b: PredicateArgument
    c: PredicateArgument
    d: PredicateArgument


class ParaPredicate(typing.NamedTuple):
    name: Literal["para"]
    A: VarName
    B: VarName
    C: VarName
    D: VarName


class EqangleStatement(typing.NamedTuple):
    name: Literal["eqangle"]
    a: PredicateArgument
    b: PredicateArgument
    c: PredicateArgument
    d: PredicateArgument
    e: PredicateArgument
    f: PredicateArgument
    g: PredicateArgument
    h: PredicateArgument


class EqanglePredicate(typing.NamedTuple):
    name: Literal["eqangle"]
    A: VarName
    B: VarName
    C: VarName
    D: VarName
    E: VarName
    F: VarName
    G: VarName
    H: VarName


class EqratioStatement(typing.NamedTuple):
    name: Literal["eqratio"]
    a: PredicateArgument
    b: PredicateArgument
    c: PredicateArgument
    d: PredicateArgument
    e: PredicateArgument
    f: PredicateArgument
    g: PredicateArgument
    h: PredicateArgument


class EqratioPredicate(typing.NamedTuple):
    name: Literal["eqratio"]
    A: VarName
    B: VarName
    C: VarName
    D: VarName
    E: VarName
    F: VarName
    G: VarName
    H: VarName


class CongStatement(typing.NamedTuple):
    name: Literal["cong"]
    a: PredicateArgument
    b: PredicateArgument
    c: PredicateArgument
    d: PredicateArgument


class CongPredicate(typing.NamedTuple):
    name: Literal["cong"]
    A: VarName
    B: VarName
    C: VarName
    D: VarName


class MidptStatement(typing.NamedTuple):
    name: Literal["midp"]
    a: PredicateArgument
    b: PredicateArgument
    c: PredicateArgument


class MidptPredicate(typing.NamedTuple):
    name: Literal["midp"]
    A: VarName
    B: VarName
    C: VarName


class TriangleStatement(typing.NamedTuple):
    name: Literal["simtri", "simtrir", "contri", "contrir"]
    a: PredicateArgument
    b: PredicateArgument
    c: PredicateArgument
    p: PredicateArgument
    q: PredicateArgument
    r: PredicateArgument


class TrianglePredicate(typing.NamedTuple):
    name: Literal["simtri", "simtrir", "contri", "contrir"]
    A: VarName
    B: VarName
    C: VarName
    P: VarName
    Q: VarName
    R: VarName


class CollStatement(tuple[str, ...]):
    def __new__(cls, predicate_name: str, *points: str):
        return super().__new__(cls, (predicate_name, *points))


class CollPredicate(tuple[str, ...]):
    def __new__(cls, predicate_name: str, *points: str):
        return super().__new__(cls, (predicate_name, *points))


class CircleStatement(tuple[str, ...]):
    def __new__(cls, predicate_name: str, *points: str):
        return super().__new__(cls, (predicate_name, *points))


class CirclePredicate(tuple[str, ...]):
    def __new__(cls, predicate_name: str, *points: str):
        return super().__new__(cls, (predicate_name, *points))


class CyclicStatement(tuple[str, ...]):
    def __new__(cls, predicate_name: str, *points: str):
        return super().__new__(cls, (predicate_name, *points))


class CyclicPredicate(tuple[str, ...]):
    def __new__(cls, predicate_name: str, *points: str):
        return super().__new__(cls, (predicate_name, *points))


EfficientStatement: TypeAlias = Union[
    PerpStatement,
    ParaStatement,
    EqangleStatement,
    EqratioStatement,
    CongStatement,
    MidptStatement,
    CollStatement,
    CircleStatement,
    TriangleStatement,
    CyclicStatement,
]

EfficientPredicateType: TypeAlias = Union[
    PerpPredicate,
    ParaPredicate,
    EqanglePredicate,
    EqratioPredicate,
    CongPredicate,
    MidptPredicate,
    CollPredicate,
    CirclePredicate,
    TrianglePredicate,
    CyclicPredicate,
]


def generate_permutations(stmt: EfficientStatement) -> Iterator[EfficientStatement]:
    match stmt:
        case PerpStatement():
            return perp_perms(stmt)
        case ParaStatement():
            return para_perms(stmt)
        case EqangleStatement():
            return eqangle_perms(stmt)
        case EqratioStatement():
            return eqratio_perms(stmt)
        case CongStatement():
            return cong_perms(stmt)
        case MidptStatement():
            return midpt_perms(stmt)
        case CollStatement():
            return coll_perms(stmt)
        case CircleStatement():
            return circle_perms(stmt)
        case TriangleStatement():
            return triangle_perms(stmt)
        case CyclicStatement():
            return cyclic_perms(stmt)
    raise ValueError(f"Unknown symmetries for statement: {stmt}")


def cyclic_perms(stmt: CyclicStatement) -> Iterator[CyclicStatement]:
    pts = stmt[1:]
    for perm in itertools.permutations(pts):
        yield CyclicStatement("cyclic", *perm)


def coll_perms(stmt: CollStatement) -> Iterator[CollStatement]:
    pts = stmt[1:]
    for perm in itertools.permutations(pts):
        yield CollStatement("coll", *perm)


def circle_perms(stmt: CircleStatement) -> Iterator[CircleStatement]:
    center, pts = stmt[1], stmt[2:]
    for perm in itertools.permutations(pts):
        yield CircleStatement("circle", center, *perm)


def triangle_perms(stmt: TriangleStatement) -> Iterator[TriangleStatement]:
    pts = stmt[1:]
    a, b, c, p, q, r = pts
    for (a1, a2, a3), (p1, p2, p3) in zip(
        itertools.permutations((a, b, c), 3),
        itertools.permutations((p, q, r), 3),
    ):
        yield TriangleStatement("simtri", a1, a2, a3, p1, p2, p3)
        yield TriangleStatement("simtri", p1, p2, p3, a1, a2, a3)


def normalize_triangle(stmt_pts: TriangleStatement) -> TriangleStatement:
    relation, a, b, c, p, q, r = stmt_pts
    (a0, p0), (b0, q0), (c0, r0) = sorted(((a, p), (b, q), (c, r)))
    (a1, p1), (b1, q1), (c1, r1) = sorted(((p, a), (q, b), (r, c)))
    return TriangleStatement(
        relation, *min((a0, b0, c0, p0, q0, r0), (a1, b1, c1, p1, q1, r1))
    )


def normalize_midpt(stmt_pts: MidptStatement) -> MidptStatement:
    _, m, a, b = stmt_pts
    return MidptStatement("midp", m, a, b)


def midpt_perms(stmt_pts: MidptStatement) -> Iterator[MidptStatement]:
    _, m, a, b = stmt_pts
    yield MidptStatement("midp", m, a, b)
    yield MidptStatement("midp", m, b, a)


def normalize_cong(stmt_pts: CongStatement) -> CongStatement:
    _, a, b, c, d = stmt_pts
    a, b = sorted((a, b))
    c, d = sorted((c, d))
    return CongStatement("cong", a, b, c, d)


def cong_perms(
    stmt_pts: CongStatement,
) -> Iterator[CongStatement]:
    _, a, b, c, d = stmt_pts
    for l1, l2 in [((a, b), (c, d)), ((c, d), (a, b))]:
        for l1_perm in itertools.permutations(l1):
            for l2_perm in itertools.permutations(l2):
                yield (
                    CongStatement(
                        "cong",
                        l1_perm[0],
                        l1_perm[1],
                        l2_perm[0],
                        l2_perm[1],
                    )
                )


def normalize_eqratio(stmt_pts: EqratioStatement) -> EqratioStatement:
    _, a, b, c, d, e, f, g, h = stmt_pts
    a, b = sorted((a, b))
    c, d = sorted((c, d))
    e, f = sorted((e, f))
    g, h = sorted((g, h))
    a, b, c, d, e, f, g, h = min(
        [
            (a, b, c, d, e, f, g, h),
            (c, d, a, b, g, h, e, f),
            (e, f, g, h, a, b, c, d),
            (g, h, e, f, c, d, a, b),
        ]
    )

    return EqratioStatement("eqratio", a, b, c, d, e, f, g, h)


def eqratio_perms(
    stmt_pts: EqratioStatement,
) -> Iterator[EqratioStatement]:
    _, a, b, c, d, e, f, g, h = stmt_pts
    for l1, l2, l3, l4 in [
        ((a, b), (c, d), (e, f), (g, h)),
        ((c, d), (a, b), (g, h), (e, f)),
        ((e, f), (g, h), (a, b), (c, d)),
        ((g, h), (e, f), (c, d), (a, b)),
    ]:
        for l1_perm in itertools.permutations(l1):
            for l2_perm in itertools.permutations(l2):
                for l3_perm in itertools.permutations(l3):
                    for l4_perm in itertools.permutations(l4):
                        yield (
                            EqratioStatement(
                                "eqratio",
                                l1_perm[0],
                                l1_perm[1],
                                l2_perm[0],
                                l2_perm[1],
                                l3_perm[0],
                                l3_perm[1],
                                l4_perm[0],
                                l4_perm[1],
                            )
                        )


def normalize_eqangle(
    stmt_pts: EqangleStatement,
) -> EqangleStatement:
    _, a, b, c, d, e, f, g, h = stmt_pts
    a, b = sorted((a, b))
    c, d = sorted((c, d))
    e, f = sorted((e, f))
    g, h = sorted((g, h))
    return EqangleStatement(
        "eqangle",
        *min(
            (a, b, c, d, e, f, g, h),
            (c, d, a, b, g, h, e, f),
            (e, f, g, h, a, b, c, d),
            (g, h, e, f, c, d, a, b),
        ),
    )


def eqangle_perms(
    stmt_pts: EqangleStatement,
) -> Iterator[EqangleStatement]:
    _, a, b, c, d, e, f, g, h = stmt_pts

    for l1, l2, l3, l4 in [
        ((a, b), (c, d), (e, f), (g, h)),
        ((c, d), (a, b), (g, h), (e, f)),
        ((e, f), (g, h), (a, b), (c, d)),
        ((g, h), (e, f), (c, d), (a, b)),
    ]:
        for l1_perm in itertools.permutations(l1):
            for l2_perm in itertools.permutations(l2):
                for l3_perm in itertools.permutations(l3):
                    for l4_perm in itertools.permutations(l4):
                        yield (
                            EqangleStatement(
                                "eqangle",
                                l1_perm[0],
                                l1_perm[1],
                                l2_perm[0],
                                l2_perm[1],
                                l3_perm[0],
                                l3_perm[1],
                                l4_perm[0],
                                l4_perm[1],
                            )
                        )


def normalize_para(stmt_pts: ParaStatement) -> ParaStatement:
    _, a, b, c, d = stmt_pts
    a, b = sorted((a, b))
    c, d = sorted((c, d))
    return ParaStatement("para", a, b, c, d)


def para_perms(
    stmt_pts: ParaStatement,
) -> Iterator[ParaStatement]:
    _, a, b, c, d = stmt_pts
    for l1, l2 in [((a, b), (c, d)), ((c, d), (a, b))]:
        for l1_perm in itertools.permutations(l1):
            for l2_perm in itertools.permutations(l2):
                yield (
                    ParaStatement(
                        "para",
                        l1_perm[0],
                        l1_perm[1],
                        l2_perm[0],
                        l2_perm[1],
                    )
                )


def normalize_perp(stmt_pts: PerpStatement) -> PerpStatement:
    _, a, b, c, d = stmt_pts
    a, b = sorted((a, b))
    c, d = sorted((c, d))
    return PerpStatement("perp", a, b, c, d)


def perp_perms(
    stmt_pts: PerpStatement,
) -> Iterator[PerpStatement]:
    _, a, b, c, d = stmt_pts
    for l1, l2 in [((a, b), (c, d)), ((c, d), (a, b))]:
        for l1_perm in itertools.permutations(l1):
            for l2_perm in itertools.permutations(l2):
                yield (
                    PerpStatement(
                        "perp",
                        l1_perm[0],
                        l1_perm[1],
                        l2_perm[0],
                        l2_perm[1],
                    )
                )


def normalize_coll(stmt_pts: CollStatement) -> CollStatement:
    pts = stmt_pts[1:]
    return CollStatement("coll", *sorted(pts))


def normalize_circle(stmt_pts: CircleStatement) -> CircleStatement:
    center, pts = stmt_pts[1], stmt_pts[2:]
    return CircleStatement("circle", center, *sorted(pts))


def normalize_cyclic(stmt_pts: CyclicStatement) -> CyclicStatement:
    pts = stmt_pts[1:]
    return CyclicStatement("cyclic", *sorted(pts))


def get_representative_of_equivalence_class(
    stmt: EfficientStatement,
) -> EfficientStatement:
    match stmt:
        case EqangleStatement():
            return normalize_eqangle(stmt)
        case EqratioStatement():
            return normalize_eqratio(stmt)
        case CongStatement():
            return normalize_cong(stmt)
        case MidptStatement():
            return normalize_midpt(stmt)
        case PerpStatement():
            return normalize_perp(stmt)
        case ParaStatement():
            return normalize_para(stmt)
        case CollStatement():
            return normalize_coll(stmt)
        case CircleStatement():
            return normalize_circle(stmt)
        case TriangleStatement():
            return normalize_triangle(stmt)
        case CyclicStatement():
            return normalize_cyclic(stmt)
    raise ValueError(f"Unknown statement to normalize: {stmt}")



import itertools
import logging
from typing import Literal

import numpy
from pydantic import BaseModel

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.jgex.clause import JGEXClause
from newclid.predicate_types import PredicateArgument

LOGGER = logging.getLogger(__name__)


class MidpointHeuristicConfig(BaseModel):
    heuristic_name: Literal[HeuristicName.MIDPOINT] = HeuristicName.MIDPOINT


class MidpointHeuristic(Heuristic):
    """Adds a number of midpoints of pairs of points present in the JGEXProblem."""

    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: numpy.random.Generator,
    ) -> list[JGEXClause]:
        pairs_of_points = list(itertools.combinations(setup.points, 2))
        added_midpoints: list[PredicateArgument] = []
        new_clauses: list[JGEXClause] = []
        for _ in range(max_new_points):
            if not pairs_of_points:
                break

            # Randomly select a pair of points to create a midpoint
            pair_index = rng.choice(len(pairs_of_points))
            p1, p2 = pairs_of_points.pop(pair_index)

            new_point = get_available_from_alphabet(
                list(setup.points) + added_midpoints
            )
            if new_point is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping midpoint addition."
                )
                break

            added_midpoints.append(new_point)

            new_clause_string = f"{new_point} = midpoint {new_point} {p1} {p2}"
            new_clauses.append(JGEXClause.from_str(new_clause_string)[0])

        return new_clauses



from newclid.ggb.read_commands import (
    GGBCommand,
    GGBMidpointCommand,
    GGBReadCommandType,
)
from newclid.ggb.read_elements import GGBConic, GGBLine
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.problem import PredicateConstruction


def commands_to_newclid_assumptions(
    commands: list[GGBCommand], lines: dict[str, GGBLine], conics: dict[str, GGBConic]
) -> list[PredicateConstruction]:
    assumptions: list[PredicateConstruction] = []
    for command in commands:
        assumptions.extend(_command_to_newclid_assumptions(command, lines, conics))
    return assumptions


def _command_to_newclid_assumptions(
    command: GGBCommand, lines: dict[str, GGBLine], conics: dict[str, GGBConic]
) -> list[PredicateConstruction]:
    match command.construction_type:
        case GGBReadCommandType.Midpoint:
            return _midpoint_command_assumption(command)
        case (
            GGBReadCommandType.Line
            | GGBReadCommandType.Intersect
            | GGBReadCommandType.AddPointOn
            | GGBReadCommandType.Circle
            | GGBReadCommandType.Circumcircle
            | GGBReadCommandType.Eqdistance
            | GGBReadCommandType.Center
            | GGBReadCommandType.OrthogonalLine
            | GGBReadCommandType.ParallelLine
            | GGBReadCommandType.AngularBisectorPoints
            | GGBReadCommandType.AngularBisector
            | GGBReadCommandType.TangentOnCircle
            | GGBReadCommandType.TangentOutsideCircle
        ):
            return []
    raise NotImplementedError(f"Unknown command: {command}")


def _midpoint_command_assumption(
    command: GGBMidpointCommand,
) -> list[PredicateConstruction]:
    return [
        PredicateConstruction.from_predicate_type_and_args(
            PredicateType.MIDPOINT,
            (
                PredicateArgument(command.midpoint),
                PredicateArgument(command.point_a),
                PredicateArgument(command.point_b),
            ),
        )
    ]



from __future__ import annotations

import re
from enum import Enum
from typing import Annotated, Literal
from xml.etree.ElementTree import Element

from pydantic import BaseModel, Field

from newclid.ggb.read_elements import GGBConic, GGBLine, GGBPoint


def read_commands(
    root: Element,
    points: dict[str, GGBPoint],
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> list[GGBCommand]:
    commands: list[GGBCommand] = []
    for xml_command in root.iter("command"):
        command = _read_command(xml_command, points, lines, conics)
        if command is not None:
            commands += command
    return commands


class GGBReadCommandType(Enum):
    Line = "Line"
    OrthogonalLine = "OrthogonalLine"
    Intersect = "Intersect"
    ParallelLine = "ParallelLine"
    Circle = "Circle"
    AddPointOn = "AddPointOn"
    Eqdistance = "Eqdistance"
    Circumcircle = "Circumcircle"
    Segment = "Segment"
    Center = "Center"
    Midpoint = "Midpoint"
    AngularBisector = "AngularBisector"
    AngularBisectorPoints = "AngularBisectorPoints"
    TangentOnCircle = "TangentOnCircle"
    TangentOutsideCircle = "TangentOutsiteCircle"


class GGBLineCommand(BaseModel):
    point_start: str
    point_end: str
    line: str

    construction_type: Literal[GGBReadCommandType.Line] = GGBReadCommandType.Line


class GGBParaCommand(BaseModel):
    point_start: str
    base_line: str
    parallel_line: str

    construction_type: Literal[GGBReadCommandType.ParallelLine] = (
        GGBReadCommandType.ParallelLine
    )


class GGBOrthogonalLineCommand(BaseModel):
    line: str
    point: str
    perpendicular_line: str

    construction_type: Literal[GGBReadCommandType.OrthogonalLine] = (
        GGBReadCommandType.OrthogonalLine
    )


class GGBIntersectCommand(BaseModel):
    object_1: str
    object_2: str
    point: str

    construction_type: Literal[GGBReadCommandType.Intersect] = (
        GGBReadCommandType.Intersect
    )


class GGBCircleCommand(BaseModel):
    center: str
    point_b: str
    circle: str
    construction_type: Literal[GGBReadCommandType.Circle] = GGBReadCommandType.Circle


class GGBCircumcircleCommand(BaseModel):
    point_a: str
    point_b: str
    point_c: str
    circle: str
    construction_type: Literal[GGBReadCommandType.Circumcircle] = (
        GGBReadCommandType.Circumcircle
    )


class GGBCompassCommand(BaseModel):
    center: str
    segment_radius: tuple[str, str]
    circle: str
    construction_type: Literal[GGBReadCommandType.Eqdistance] = (
        GGBReadCommandType.Eqdistance
    )


class GGBAddPointOnObjectCommand(BaseModel):
    object_name: str
    point: str
    construction_type: Literal[GGBReadCommandType.AddPointOn] = (
        GGBReadCommandType.AddPointOn
    )


class GGBCenterCommand(BaseModel):
    circle: str
    point: str
    construction_type: Literal[GGBReadCommandType.Center] = GGBReadCommandType.Center


class GGBMidpointCommand(BaseModel):
    point_a: str
    point_b: str
    midpoint: str
    construction_type: Literal[GGBReadCommandType.Midpoint] = (
        GGBReadCommandType.Midpoint
    )


class GGBAngularBisectorLinesCommand(BaseModel):
    line_1: str
    line_2: str
    line_bisector: str
    construction_type: Literal[GGBReadCommandType.AngularBisector] = (
        GGBReadCommandType.AngularBisector
    )


class GGBAngularBisectorPointsCommand(BaseModel):
    point_a: str
    corner: str
    point_b: str
    line_bisector: str
    construction_type: Literal[GGBReadCommandType.AngularBisectorPoints] = (
        GGBReadCommandType.AngularBisectorPoints
    )


class GGBTangentOnCircleCommand(BaseModel):
    point_on_tangent: str
    tangent_object: str
    tangent_line: str
    construction_type: Literal[GGBReadCommandType.TangentOnCircle] = (
        GGBReadCommandType.TangentOnCircle
    )


class GGBTangentOutsideCircleCommand(BaseModel):
    point_on_tangent: str
    tangent_object: str
    tangent_line1: str
    tangent_line2: str
    construction_type: Literal[GGBReadCommandType.TangentOutsideCircle] = (
        GGBReadCommandType.TangentOutsideCircle
    )


GGBCommand = Annotated[
    GGBOrthogonalLineCommand
    | GGBLineCommand
    | GGBIntersectCommand
    | GGBParaCommand
    | GGBCircleCommand
    | GGBAddPointOnObjectCommand
    | GGBCircumcircleCommand
    | GGBCompassCommand
    | GGBCenterCommand
    | GGBMidpointCommand
    | GGBAngularBisectorLinesCommand
    | GGBTangentOnCircleCommand
    | GGBTangentOutsideCircleCommand
    | GGBAngularBisectorPointsCommand,
    Field(discriminator="construction_type"),
]


class GGBXMLCommandType(Enum):
    Circle = "Circle"
    Center = "Center"
    Line = "Line"
    Segment = "Segment"
    OrthogonalLine = "OrthogonalLine"
    Point = "Point"
    Intersect = "Intersect"
    Midpoint = "Midpoint"
    AngularBisector = "AngularBisector"
    Tangent = "Tangent"


def _read_command(
    command: Element,
    points: dict[str, GGBPoint],
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> list[GGBCommand] | None:
    command_name = GGBXMLCommandType(command.attrib["name"])
    inputs, outputs = list(command)
    match command_name:
        case GGBXMLCommandType.Line:
            return _read_line_xml(inputs, outputs, points, lines)
        case GGBXMLCommandType.OrthogonalLine:
            return [
                GGBOrthogonalLineCommand(
                    point=inputs.attrib["a0"],
                    line=inputs.attrib["a1"],
                    perpendicular_line=outputs.attrib["a0"],
                )
            ]
        case GGBXMLCommandType.Circle:
            return _read_circle_xml(inputs, outputs, points)
        case (
            GGBXMLCommandType.Point
            | GGBXMLCommandType.Intersect
            | GGBXMLCommandType.Center
            | GGBXMLCommandType.Midpoint
        ):
            return _read_point_creation_xml(command_name, inputs, outputs, points)
        case GGBXMLCommandType.AngularBisector:
            return _read_angular_bisector_xml(inputs, outputs, points, lines)
        case GGBXMLCommandType.Tangent:
            return _read_tangent_xml(inputs, outputs, points, lines, conics)
        case GGBXMLCommandType.Segment:
            return None
    raise NotImplementedError(f"Unknown command: {command_name}")


def _read_point_creation_xml(
    command_name: GGBXMLCommandType,
    inputs: Element,
    outputs: Element,
    points: dict[str, GGBPoint],
) -> list[GGBCommand]:
    point_name = outputs.attrib["a0"]
    if point_name not in points:
        # Skip hidden points
        return []
    match command_name:
        case GGBXMLCommandType.Point:
            return [
                GGBAddPointOnObjectCommand(
                    object_name=inputs.attrib["a0"],
                    point=outputs.attrib["a0"],
                )
            ]
        case GGBXMLCommandType.Intersect:
            return [
                GGBIntersectCommand(
                    object_1=inputs.attrib["a0"],
                    object_2=inputs.attrib["a1"],
                    point=outputs.attrib["a0"],
                )
            ]
        case GGBXMLCommandType.Midpoint:
            if outputs.attrib["a0"] in points:
                return [
                    GGBMidpointCommand(
                        point_a=inputs.attrib["a0"],
                        point_b=inputs.attrib["a1"],
                        midpoint=outputs.attrib["a0"],
                    )
                ]
        case GGBXMLCommandType.Center:
            return [
                GGBCenterCommand(
                    circle=inputs.attrib["a0"],
                    point=outputs.attrib["a0"],
                )
            ]
        case _:  # pragma: no cover
            pass
    raise NotImplementedError(f"Unknown point creation command: {command_name}")


def _read_line_xml(
    inputs: Element,
    outputs: Element,
    points: dict[str, GGBPoint],
    lines: dict[str, GGBLine],
) -> list[GGBCommand]:
    point_or_line = inputs.attrib["a1"]
    if point_or_line in points:
        return [
            GGBLineCommand(
                point_start=inputs.attrib["a0"],
                point_end=point_or_line,
                line=outputs.attrib["a0"],
            )
        ]
    if point_or_line in lines:
        return [
            GGBParaCommand(
                point_start=inputs.attrib["a0"],
                base_line=point_or_line,
                parallel_line=outputs.attrib["a0"],
            )
        ]
    raise NotImplementedError(f"Unknown line object: {point_or_line}")


def _read_circle_xml(
    inputs: Element, outputs: Element, points: dict[str, GGBPoint]
) -> list[GGBCommand]:
    point_or_segment = inputs.attrib["a1"]
    if point_or_segment in points:
        if "a2" in inputs.attrib:
            return [
                GGBCircumcircleCommand(
                    point_a=inputs.attrib["a0"],
                    point_b=point_or_segment,
                    point_c=inputs.attrib["a2"],
                    circle=outputs.attrib["a0"],
                )
            ]
        else:
            return [
                GGBCircleCommand(
                    center=inputs.attrib["a0"],
                    point_b=point_or_segment,
                    circle=outputs.attrib["a0"],
                )
            ]
    else:
        radius_extrema_components = _str_to_ggbcommand_components(point_or_segment)
        if len(radius_extrema_components.args) != 2:  # pragma: no cover
            raise ValueError(
                f"Expected 2 arguments for radius extrema, got {radius_extrema_components.args}"
            )
        a, b = radius_extrema_components.args
        return [
            GGBCompassCommand(
                center=inputs.attrib["a0"],
                segment_radius=(a, b),
                circle=outputs.attrib["a0"],
            )
        ]


def _read_angular_bisector_xml(
    inputs: Element,
    outputs: Element,
    points: dict[str, GGBPoint],
    lines: dict[str, GGBLine],
) -> list[GGBCommand]:
    arg_1 = inputs.attrib["a0"]
    if arg_1 in lines:
        return [
            GGBAngularBisectorLinesCommand(
                line_1=arg_1,
                line_2=inputs.attrib["a1"],
                line_bisector=outputs.attrib["a0"],
            ),
            GGBAngularBisectorLinesCommand(
                line_1=arg_1,
                line_2=inputs.attrib["a1"],
                line_bisector=outputs.attrib["a1"],
            ),
        ]
    elif arg_1 in points:
        return [
            GGBAngularBisectorPointsCommand(
                point_a=arg_1,
                corner=inputs.attrib["a1"],
                point_b=inputs.attrib["a2"],
                line_bisector=outputs.attrib["a0"],
            )
        ]
    raise NotImplementedError(f"Unknown angular bisector object: {arg_1}")


def _read_tangent_xml(
    inputs: Element,
    outputs: Element,
    points: dict[str, GGBPoint],
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> list[GGBCommand]:
    arg_1 = inputs.attrib["a0"]
    arg_2 = inputs.attrib["a1"]
    if arg_1 in points and arg_2 in conics:
        if not outputs.attrib["a1"]:
            return [
                GGBTangentOnCircleCommand(
                    point_on_tangent=arg_1,
                    tangent_object=arg_2,
                    tangent_line=outputs.attrib["a0"],
                )
            ]
        else:
            return [
                GGBTangentOutsideCircleCommand(
                    point_on_tangent=arg_1,
                    tangent_object=arg_2,
                    tangent_line1=outputs.attrib["a0"],
                    tangent_line2=outputs.attrib["a1"],
                )
            ]
    raise NotImplementedError("This situation is not handled yet.")


class GGBXMLCommandComponents(BaseModel):
    command_type: str
    args: list[str]


def _str_to_ggbcommand_components(command_str: str) -> GGBXMLCommandComponents:
    split_string = re.match(r"(\w+)\[(.*)\]", command_str)

    if not split_string:  # pragma: no cover
        raise ValueError(f"This may not be a command: '{command_str}'")

    command_type = split_string.group(1)
    command_args = split_string.group(2).split(", ")

    return GGBXMLCommandComponents(command_type=command_type, args=command_args)



import logging

import numpy as np
import pytest
from newclid.all_rules import (
    R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
    R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES,
    R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES,
    R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES,
)
from newclid.api import GeometricSolverBuilder
from newclid.jgex.problem_builder import JGEXProblemBuilder

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


TEST_CASES = [
    (
        "simtri_no_rep",
        R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
        "a b c = triangle a b c; p q = segment p q; r = simtri r a b c p q ? eqangle b a b c q p q r; eqangle a b a c p q p r; eqangle a c b c p r q r; eqratio b a b c q p q r; eqratio b c a c q r p r",
    ),
    (
        "simtri_a=p",
        R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
        "a b c = triangle a b c; q = free q; r = simtri r a b c a q ? eqangle b a b c q a q r; eqangle a b a c a q a r; eqangle a c b c a r q r; eqratio b a b c q a q r; eqratio b c a c q r a r",
    ),
    (
        "simtri_a=q",
        R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
        "a b c = triangle a b c; p =free p; r = simtri r a b c p a ? eqangle b a b c a p a r; eqangle a b a c p a p r; eqangle a c b c p r a r; eqratio b a b c a p a r; eqratio b c a c a r p r",
    ),
    (
        "simtri_a=r",
        R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
        "a b c = triangle a b c; q = free q; p = simtri p b c a q a ? eqangle b a b c q p q a; eqangle a b a c p q p a; eqangle a c b c p a q a; eqratio b a b c q p q a; eqratio b c a c q a p a",
    ),
    (
        "simtri_a=q_b=r",
        R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
        "a b c = triangle a b c; p = simtri p b c a a b ? eqangle b a b c a p a b; eqangle a b a c p a p b; eqangle a c b c p b a b; eqratio b a b c a p a b; eqratio b c a c a b p b",
    ),
    (
        "simtrir_no_rep",
        R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES,
        "a b c = triangle a b c; p q = segment p q; r = simtrir r a b c p q ? eqangle b a b c q r q p; eqangle a b a c p r p q; eqangle a c b c q r p r; eqratio b a b c q p q r; eqratio b c a c q r p r",
    ),
    (
        "simtrir_a=p",
        R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES,
        "a b c = triangle a b c; q = free q; r = simtrir r a b c a q ? eqangle b a b c q r q a; eqangle a b a c a r a q; eqangle a c b c q r a r; eqratio b a b c q a q r; eqratio b c a c q r a r",
    ),
    (
        "simtrir_a=p_b=q",
        R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES,
        "a b c = triangle a b c; r = simtrir r a b c a b ? eqangle b a b c b r b a; eqangle a b a c a r a b; eqangle a c b c b r a r; eqratio b a b c b a b r; eqratio b c a c b r a r",
    ),
    (
        "contri_no_rep",
        R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES,
        "a b c = triangle a b c; p = free p; q r = contri q r a b c p ? simtri a b c p q r; cong a b p q",
    ),
    (
        "contri_a=p",
        R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES,
        "a b c = triangle a b c; q r = contri q r a b c a ? simtri a b c a q r; cong a b a q",
    ),
    (
        "contrir_no_rep",
        R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES,
        "a b c = triangle a b c; p = free p; q r = contrir q r a b c p ? simtrir a b c p q r; cong a b p q",
    ),
    (
        "contrir_a=p",
        R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES,
        "a b c = triangle a b c; q r = contrir q r a b c a ? simtrir a b c a q r; cong a b a q",
    ),
]

TEST_CASE_NAMES = [rule[0] for rule in TEST_CASES]
assert len(TEST_CASE_NAMES) == len(set(TEST_CASE_NAMES)), (
    "There are duplicate case names"
)


@pytest.mark.parametrize("case_name", TEST_CASE_NAMES)
def test_triangle_case(case_name: str):
    _, rule, problem_txt = TEST_CASES[TEST_CASE_NAMES.index(case_name)]
    LOGGER.info(f"Testing case {case_name} with problem {problem_txt} and rule {rule}")
    rng = np.random.default_rng(123)
    solver_builder = GeometricSolverBuilder(rng=rng).with_rules([rule])
    problem_builder = JGEXProblemBuilder(rng=rng).with_problem_from_txt(problem_txt)
    solver = solver_builder.build(problem_builder.build())
    success = solver.run()
    assert success



"""Interface for the defaults of the API given installed plugins."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from newclid.api import DeductiveAgent, Deductor, RuleMatcher
    from newclid.proof_state import ProofState


class APIDefault(ABC):
    @abstractmethod
    def default_rule_matcher(self) -> RuleMatcher:
        """Return the default rule matcher for the API."""

    @abstractmethod
    def default_deductors(self) -> list[Deductor]:
        """Return the default deductors for the API."""

    @abstractmethod
    def default_deductive_agent(self) -> DeductiveAgent:
        """Return the default deductive agent for the API."""

    @abstractmethod
    def callback(self, proof_state: ProofState) -> None:
        """Callback function to be called after each step of the proof."""



import logging
from pathlib import Path
from typing import Self

from numpy.random import Generator as RngGenerator

from newclid.api import ProblemBuilder
from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions import ALL_JGEX_CONSTRUCTIONS
from newclid.jgex.definition import JGEXDefinition
from newclid.jgex.errors import JGEXConstructionError
from newclid.jgex.formulation import (
    JGEXFormulation,
    alphabetize,
    jgex_formulation_from_txt_file,
)
from newclid.jgex.to_newclid import JGEXClauseConsequences, build_newclid_problem
from newclid.problem import ProblemSetup, nc_problem_is_valid
from newclid.rng import setup_rng

LOGGER = logging.getLogger(__name__)


class JGEXProblemBuilder(ProblemBuilder):
    def __init__(
        self,
        rng: RngGenerator | int | None = None,
        problem: JGEXFormulation | None = None,
        max_attempts_per_clause: int = 5,
    ) -> None:
        self.jgex_problem = problem
        self.rng = setup_rng(rng)
        self.max_attempts_per_clause = max_attempts_per_clause
        self.jgex_defs: dict[str, JGEXDefinition] = JGEXDefinition.to_dict(
            ALL_JGEX_CONSTRUCTIONS
        )
        self._clauses_consequences: dict[JGEXClause, JGEXClauseConsequences] | None = (
            None
        )
        self._include_auxiliary_clauses: bool = False

    def build(
        self, max_attempts_to_satisfy_goals_numerically: int = 100
    ) -> ProblemSetup:
        if self.jgex_problem is None:
            raise ValueError("JGEX problem is not set by any means.")

        valid_problem_setup = None
        last_error = JGEXConstructionError(
            f"Did not even try to build a valid problem setup for {self.jgex_problem.name}."
        )
        for _attempt in range(max_attempts_to_satisfy_goals_numerically):
            try:
                problem_setup, clauses_consequences = build_newclid_problem(
                    problem=self.jgex_problem,
                    defs=self.jgex_defs,
                    max_attempts_per_clause=self.max_attempts_per_clause,
                    rng=self.rng,
                    include_auxiliary_clauses=self._include_auxiliary_clauses,
                )
                if nc_problem_is_valid(problem_setup):
                    valid_problem_setup = problem_setup
                    self._clauses_consequences = clauses_consequences
                    break
            except JGEXConstructionError as e:
                LOGGER.debug(
                    f"Failed to build a valid problem setup (attempt {_attempt}): {e}"
                )
                last_error = e
                continue

        if valid_problem_setup is None:
            raise ValueError(
                f"Failed to build a valid problem setup for {self.jgex_problem.name} after {max_attempts_to_satisfy_goals_numerically} attempts. Last error: {last_error}"
            ) from last_error

        return valid_problem_setup

    def with_problem(self, problem: JGEXFormulation) -> Self:
        self.jgex_problem = problem
        return self

    def with_problem_from_txt(
        self, problem_txt: str, problem_name: str = "problem"
    ) -> Self:
        self.jgex_problem = JGEXFormulation.from_text(problem_txt)
        self.jgex_problem.name = problem_name
        return self

    def include_auxiliary_clauses(self, include: bool = True) -> Self:
        self._include_auxiliary_clauses = include
        return self

    def with_problem_from_file(
        self, problems_path: Path, problem_name: str, rename: bool = False
    ) -> Self:
        problems = jgex_formulation_from_txt_file(problems_path)
        try:
            self.jgex_problem = problems[problem_name]
        except KeyError as e:
            raise KeyError(f"{problem_name} not found in file {problems_path}") from e
        if rename:
            self.jgex_problem, _reversed_mapping = alphabetize(self.jgex_problem)
        return self

    def with_defs(self, defs: list[JGEXDefinition]) -> Self:
        self.jgex_defs = JGEXDefinition.to_dict(defs)
        return self

    @property
    def clauses_consequences(self) -> dict[JGEXClause, JGEXClauseConsequences]:
        if self._clauses_consequences is None:
            raise ValueError(
                "JGEX clauses consequences are not set."
                " You need to build first to use this property."
            )
        return self._clauses_consequences



"""Definitions of free points in a specific formation or figure."""

from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.definition import JGEXDefinition, SketchConstruction, v
from newclid.jgex.sketch import (
    sketch_acute_triangle,
    sketch_between,
    sketch_eq_quadrangle,
    sketch_eqdia_quadrangle,
    sketch_free,
    sketch_function_name,
    sketch_ieq_triangle,
    sketch_iso_trapezoid,
    sketch_isos,
    sketch_isquare,
    sketch_pentagon,
    sketch_quadrangle,
    sketch_r_trapezoid,
    sketch_r_triangle,
    sketch_rectangle,
    sketch_risos,
    sketch_segment,
    sketch_trapezoid,
    sketch_triangle,
    sketch_triangle12,
)

# 1 Point
FREE = JGEXDefinition(
    name=JGEXConstructionName.FREE,
    args=(v("a"),),
    rely_on_points={v("a"): ()},
    requirements=JGEXClause.from_str("")[0],
    clauses=(),
    sketches=(
        SketchConstruction(name=sketch_function_name(sketch_free.__name__), args=()),
    ),
    input_points=(),
    output_points=(v("a"),),
)

# 2 Points
SEGMENT = JGEXDefinition(
    name=JGEXConstructionName.SEGMENT,
    args=(v("a"), v("b")),
    rely_on_points={},
    requirements=JGEXClause.from_str("")[0],
    clauses=(),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_segment.__name__),
            args=(),
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b")),
)

BETWEEN = JGEXDefinition(
    name=JGEXConstructionName.BETWEEN,
    args=(v("c"), v("a"), v("b")),
    rely_on_points={},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("coll a b c, obtuse_angle a c b"),
    sketches=(
        SketchConstruction(name=sketch_function_name(sketch_between.__name__), args=()),
    ),
    input_points=(),
    output_points=(v("c"), v("a"), v("b")),
)


# 3 Points
TRIANGLE = JGEXDefinition(
    name=JGEXConstructionName.TRIANGLE,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={},
    requirements=JGEXClause.from_str("")[0],
    clauses=(),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_triangle.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)

R_TRIANGLE = JGEXDefinition(
    name=JGEXConstructionName.R_TRIANGLE,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={v("c"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("perp a b a c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_r_triangle.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)

ACUTE_TRIANGLE = JGEXDefinition(
    name=JGEXConstructionName.ACUTE_TRIANGLE,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={},
    requirements=JGEXClause.from_str("")[0],
    clauses=(),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_acute_triangle.__name__),
            args=(),
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)


ISO_TRIANGLE = JGEXDefinition(
    name=JGEXConstructionName.ISO_TRIANGLE,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={v("c"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("eqangle b a b c c b c a, cong a b a c"),
    sketches=(
        SketchConstruction(name=sketch_function_name(sketch_isos.__name__), args=()),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)


ISO_TRIANGLE0 = JGEXDefinition(
    name=JGEXConstructionName.ISO_TRIANGLE0,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={v("c"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("cong a b a c"),
    sketches=(
        SketchConstruction(name=sketch_function_name(sketch_isos.__name__), args=()),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)


IEQ_TRIANGLE = JGEXDefinition(
    name=JGEXConstructionName.IEQ_TRIANGLE,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={v("c"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str(
        "cong a b b c, cong b c c a; eqangle a b a c c a c b, eqangle c a c b b c b a"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_ieq_triangle.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)

TRIANGLE12 = JGEXDefinition(
    name=JGEXConstructionName.TRIANGLE12,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={v("c"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("rconst a b a c 1/2"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_triangle12.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)

RISOS = JGEXDefinition(
    name=JGEXConstructionName.RISOS,
    args=(v("a"), v("b"), v("c")),
    rely_on_points={v("c"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("perp a b a c, cong a b a c; eqangle b a b c c b c a"),
    sketches=(
        SketchConstruction(name=sketch_function_name(sketch_risos.__name__), args=()),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c")),
)

TRIANGLE_CONSTRUCTIONS = [
    TRIANGLE,
    R_TRIANGLE,
    ACUTE_TRIANGLE,
    ISO_TRIANGLE,
    ISO_TRIANGLE0,
    IEQ_TRIANGLE,
    TRIANGLE12,
    RISOS,
]

# 4 Points
QUADRANGLE = JGEXDefinition(
    name=JGEXConstructionName.QUADRANGLE,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={},
    requirements=JGEXClause.from_str("")[0],
    clauses=(),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_quadrangle.__name__),
            args=(),
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)


RECTANGLE = JGEXDefinition(
    name=JGEXConstructionName.RECTANGLE,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("c"): (v("a"), v("b")), v("d"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str(
        "c : perp a b b c ; d : para a b c d, para a d b c; perp a b a d, cong a b c d, cong a d b c, cong a c b d"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_rectangle.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)


ISQUARE = JGEXDefinition(
    name=JGEXConstructionName.ISQUARE,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("c"): (v("a"), v("b")), v("d"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str(
        "c : perp a b b c, cong a b b c; d : para a b c d, para a d b c; perp a d d c, cong b c c d, cong c d d a, perp a c b d, cong a c b d"
    ),
    sketches=(
        SketchConstruction(name=sketch_function_name(sketch_isquare.__name__), args=()),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)


TRAPEZOID = JGEXDefinition(
    name=JGEXConstructionName.TRAPEZOID,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("d"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("para a b c d"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_trapezoid.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)

R_TRAPEZOID = JGEXDefinition(
    name=JGEXConstructionName.R_TRAPEZOID,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("d"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("para a b c d, perp a b a d"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_r_trapezoid.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)


ISO_TRAPEZOID = JGEXDefinition(
    name=JGEXConstructionName.ISO_TRAPEZOID,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("d"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("para d c a b, cong d a b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_iso_trapezoid.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)


EQ_QUADRANGLE = JGEXDefinition(
    name=JGEXConstructionName.EQ_QUADRANGLE,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("d"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("cong d a b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_eq_quadrangle.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)

EQDIA_QUADRANGLE = JGEXDefinition(
    name=JGEXConstructionName.EQDIA_QUADRANGLE,
    args=(v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("d"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("cong d b a c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_eqdia_quadrangle.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d")),
)

QUADRANGLE_CONSTRUCTIONS = [
    QUADRANGLE,
    RECTANGLE,
    ISQUARE,
    TRAPEZOID,
    R_TRAPEZOID,
    ISO_TRAPEZOID,
    EQ_QUADRANGLE,
    EQDIA_QUADRANGLE,
]


# 5 Points
PENTAGON = JGEXDefinition(
    name=JGEXConstructionName.PENTAGON,
    args=(v("a"), v("b"), v("c"), v("d"), v("e")),
    rely_on_points={},
    requirements=JGEXClause.from_str("")[0],
    clauses=(),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_pentagon.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d"), v("e")),
)


FREE_CONSTRUCTIONS = (
    [FREE, SEGMENT, BETWEEN, PENTAGON]
    + TRIANGLE_CONSTRUCTIONS
    + QUADRANGLE_CONSTRUCTIONS
)



"""Define the Rule as used in Newclid.

Rules represent theorems of geometry that are considered as axioms in the proof.
"""

from __future__ import annotations

from pathlib import Path
from typing import Literal, NewType, Self

import yaml
from pydantic import BaseModel, ConfigDict, model_validator

from newclid.justifications._index import JustificationType
from newclid.predicates import Predicate
from newclid.tools import atomize

VarName = NewType("VarName", str)


class RuleConstruction(BaseModel):
    """A construction in a rule."""

    name: str
    """Name of the construction."""
    variables: tuple[VarName, ...]
    """Input variable names."""


class RuleApplication(BaseModel):
    """A rule application."""

    predicate: Predicate
    """The predicate justified by the rule."""

    rule: Rule
    """The rule applied."""
    premises: tuple[Predicate, ...]
    """The premises of the rule."""

    dependency_type: Literal[JustificationType.RULE_APPLICATION] = (
        JustificationType.RULE_APPLICATION
    )

    @model_validator(mode="after")
    def canonicalize(self) -> Self:
        self.premises = tuple(sorted(set(self.premises), key=lambda x: repr(x)))
        return self

    def __str__(self) -> str:
        premises_txt = ", ".join(str(premise) for premise in self.premises)
        return f"{premises_txt} =({self.rule.id})> {self.predicate}"

    def __hash__(self) -> int:
        return hash(self.predicate)


class Rule(BaseModel):
    """Deduction rule."""

    model_config = ConfigDict(frozen=True)

    id: str
    """Unique permanent identifier for the rule."""
    description: str
    """Descriptive name for the rule."""
    premises_txt: tuple[str, ...]
    """Tuple of predicates representing the hypothesis of the rule."""
    conclusions_txt: tuple[str, ...]
    """Tuple of predicates representing the conclusions of the rule."""
    allow_point_repetition: bool = False
    """Whether the same point can correspond to different variables in the rule presentation."""

    @property
    def premises(self) -> list[RuleConstruction]:
        premises: list[RuleConstruction] = []
        for p in self.premises_txt:
            name, *args = atomize(p)
            premises.append(
                RuleConstruction(name=name, variables=tuple(VarName(a) for a in args))
            )
        return premises

    @property
    def conclusions(self) -> list[RuleConstruction]:
        conclusions: list[RuleConstruction] = []
        for c in self.conclusions_txt:
            name, *args = atomize(c)
            conclusions.append(
                RuleConstruction(name=name, variables=tuple(VarName(a) for a in args))
            )
        return conclusions

    @property
    def fullname(self) -> str:
        return f"{self.id} {self.description}"

    @property
    def variables(self) -> list[VarName]:
        variable_names: set[VarName] = set()
        for p in self.premises + self.conclusions:
            for x in p.variables:
                if str.isalpha(x[0]):
                    variable_names.add(x)
        return list(variable_names)

    def __str__(self) -> str:
        premises_txt = ", ".join(premise for premise in self.premises_txt)
        conclusions_txt = ", ".join(conclusion for conclusion in self.conclusions_txt)
        return f"{premises_txt} => {conclusions_txt}"

    def __hash__(self) -> int:
        return hash(self.id)


def rules_from_file(file_path: Path) -> list[Rule]:
    """Load deduction rule from a file."""
    match file_path.suffix:
        case ".yaml":
            return _rules_from_yaml(file_path.read_text())
        case ".txt":
            return rules_from_txt(file_path.read_text())
        case _:
            raise ValueError(f"Unsupported file extension: {file_path.suffix}")


def _rules_from_yaml(yaml_content: str) -> list[Rule]:
    """Load deduction rule from a yaml object."""
    return [Rule.model_validate(rule) for rule in yaml.safe_load(yaml_content)]


def rules_from_txt(text: str) -> list[Rule]:
    """Load deduction rule from a txt object."""
    description = ""
    res: list[Rule] = []
    for i, s in enumerate(atomize(text, "\n")):
        if "=>" in s:
            res.append(_rule_from_string(s, description or f"rule of line {i}"))
            description = ""
        else:
            description = s
    return res


def _rule_from_string(s: str, description: str = "") -> Rule:
    premises, conclusions = atomize(s, "=>")
    id, description = description.split(" ", 1)
    return Rule(
        id=id,
        description=description,
        premises_txt=atomize(premises, ","),
        conclusions_txt=atomize(conclusions, ","),
    )



"""Classical Breadth-First Search based agents."""

from __future__ import annotations

import logging
import time
from collections import defaultdict
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from newclid.agent.agents_interface import DeductiveAgent
from newclid.predicates._index import PredicateType

if TYPE_CHECKING:
    from newclid.justifications.justification import Justification
    from newclid.proof_state import ProofState
    from newclid.rule import Rule

LOGGER = logging.getLogger(__name__)


class DDARNStats(BaseModel):
    """Statistics for the DDARN agent."""

    agent_type: Literal["ddarn"] = "ddarn"

    max_level: int
    """The maximum reached level of the DDARN agent."""

    level_predicate_count: list[tuple[int, PredicateType, int]]
    """How many predicates of each type do we have at each DDARN level of the diagram."""


class DDARN(DeductiveAgent):
    """Apply Deductive Derivation to exhaustion by Breadth-First Search.

    DDARN will match and apply all available rules level by level
    until reaching a fixpoint we call exhaustion.

    """

    def __init__(self) -> None:
        self.rule_buffer: list[Rule] = []
        self.exausted = False
        self.level = 0
        self.level_timer = time.time()
        self.proven_by_levels: dict[int, list[Justification]] = defaultdict(list)
        self.dd_cumulative_time = 0.0
        self.deductors_cumulative_time: dict[str, float] = defaultdict(lambda: 0.0)

    def get_stats(self) -> DDARNStats:
        return DDARNStats(
            max_level=self.level,
            level_predicate_count=_predicates_by_level(self.proven_by_levels),
        )

    def step(self, proof: ProofState, rules: list[Rule]) -> bool:
        if proof.check_goals():
            return False
        if self.rule_buffer:
            next_rule = self.rule_buffer.pop(0)
            self._match_and_apply_rule(next_rule, proof=proof)
        else:
            self._wrap_up_level(proof)
            if self.exausted:
                return False
            self._new_level(rules)
        return True

    def _match_and_apply_rule(self, rule: Rule, proof: ProofState) -> None:
        for dep in proof.match_theorem(rule):
            self._apply_dep(proof, dep)

    def _get_and_apply_from_deductors(self, proof: ProofState) -> dict[str, float]:
        time_per_deductor: dict[str, float] = {}
        for deductor in proof.deductors:
            t0 = time.time()
            for dep in deductor.deduce(proof.symbols):
                self._apply_dep(proof, dep)
            deductor_time = time.time() - t0
            time_per_deductor[deductor.__class__.__name__] = deductor_time
            self.deductors_cumulative_time[deductor.__class__.__name__] += deductor_time
        return time_per_deductor

    def _apply_dep(self, proof: ProofState, dep: Justification) -> None:
        if proof.apply(dep):
            self.proven_by_levels[self.level].append(dep)
            self.exausted = False

    def _wrap_up_level(self, proof: ProofState) -> None:
        dd_level_time = time.time() - self.level_timer
        deductors_t0 = time.time()
        time_per_deductor = self._get_and_apply_from_deductors(proof=proof)
        deductors_time = time.time() - deductors_t0
        times_str = ", ".join(
            f"{deductor_name}: {time:.2f}s"
            for deductor_name, time in time_per_deductor.items()
        )
        LOGGER.info(
            f"Level {self.level} completed in {dd_level_time + deductors_time:.2f}s"
            f" (DD: {dd_level_time:.2f}s, {times_str})"
        )
        self.dd_cumulative_time += dd_level_time

    def _new_level(self, rules: list[Rule]) -> None:
        self.exausted = True
        self.rule_buffer = list(rules)
        self.level += 1
        self.level_timer = time.time()


def _predicates_by_level(
    deps_by_level: dict[int, list[Justification]],
) -> list[tuple[int, PredicateType, int]]:
    histogram: list[tuple[int, PredicateType, int]] = []
    for level, deps in deps_by_level.items():
        level_histogram: dict[PredicateType, int] = {}
        for dep in deps:
            predicate_type = dep.predicate.predicate_type
            if predicate_type not in level_histogram:
                level_histogram[predicate_type] = 0
            level_histogram[predicate_type] += 1
        histogram.extend(
            [(level, pred, count) for pred, count in level_histogram.items()]
        )
    return histogram



import numpy as np
import pytest
from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.heuristics.apply_heuristics import (
    apply_complete_the_picture_heuristics,
    apply_heuristics_on_nc_problem,
    build_nc_problem_from_jgex_problem,
)
from newclid.heuristics.centers_of_cyclic import CentersOfCyclicHeuristic
from newclid.heuristics.foots_on_lines import FootHeuristic
from newclid.heuristics.heuristic_from_config import heuristic_from_name
from newclid.heuristics.line_intersections import LineIntersectionsHeuristic
from newclid.heuristics.midpoint import MidpointHeuristic
from newclid.heuristics.transfer_distances import TransferDistancesHeuristic
from newclid.jgex.clause import JGEXClause
from newclid.jgex.formulation import JGEXFormulation
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.problem import PredicateConstruction, ProblemSetup
from newclid.tools import pretty_basemodel_list_diff


class TestBaseCasesOfHeuristics:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = HeuristicFixture()

    @pytest.mark.parametrize("heuristic_name", HeuristicName)
    def test_each_heuristic_can_be_applied(self, heuristic_name: HeuristicName):
        """Test that each heuristic can be applied."""
        self.fixture.given_jgex_problem(
            "a b c = triangle a b c; d e = segment d e; f = on_circum f a b c;"
            " g = on_aline0 g a b c d e f d ? eqangle f a a b f c c b"
        )
        self.fixture.when_applying_heuristic(heuristic_from_name(heuristic_name))
        self.fixture.then_there_should_be_at_least_n_new_points(1)

    def test_max_points_limit(self):
        """Test that max_new_points limit is respected."""
        self.fixture.given_jgex_problem("a b c = triangle a b c")
        max_new_points = 2
        self.fixture.when_applying_heuristic(MidpointHeuristic(), max_new_points)
        self.fixture.then_there_should_be_at_least_n_new_points(1)
        self.fixture.then_there_should_be_at_most_n_new_points(max_new_points)

    def test_deterministic_behavior(self):
        """Test that heuristics produce deterministic results with same seed."""
        rng = np.random.default_rng(123)
        jgex_problem = JGEXFormulation.from_text("a b c = triangle a b c")
        nc_problem = build_nc_problem_from_jgex_problem(jgex_problem, rng)

        # Apply heuristics with same seed
        rng1 = np.random.default_rng(123)
        ncproblem1, added_clauses1 = apply_heuristics_on_nc_problem(
            nc_problem, jgex_problem, MidpointHeuristic(), rng1, max_new_points=3
        )

        rng2 = np.random.default_rng(123)
        ncproblem2, added_clauses2 = apply_heuristics_on_nc_problem(
            nc_problem, jgex_problem, MidpointHeuristic(), rng2, max_new_points=3
        )

        # Results should be identical
        assert ncproblem1.points == ncproblem2.points
        assert ncproblem1.assumptions == ncproblem2.assumptions
        assert added_clauses1 == added_clauses2


class TestLineIntersectionsHeuristic:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = HeuristicFixture()

    def test_successful_line_intersections_heuristic(self):
        """Test that line intersections heuristic works."""
        self.fixture.given_jgex_problem("a b = segment a b; c d = segment c d")
        self.fixture.when_applying_heuristic(LineIntersectionsHeuristic())
        self.fixture.then_there_should_be_new_assumptions_of_predicate(
            PredicateType.COLLINEAR
        )
        self.fixture.then_there_should_be_at_least_n_new_points(1)


class TestTransferDistancesHeuristic:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = HeuristicFixture()

    def test_successful_transfer_distances_heuristic(self):
        """Test that transfer distances heuristic works."""
        self.fixture.given_jgex_problem("a b c = triangle a b c")
        self.fixture.when_applying_heuristic(TransferDistancesHeuristic())
        self.fixture.then_there_should_be_new_assumptions_of_predicate(
            PredicateType.CONGRUENT
        )
        self.fixture.then_there_should_be_at_least_n_new_points(1)


class TestMidpointHeuristic:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = HeuristicFixture()

    def test_successful_midpoint_heuristic(self):
        """Test that midpoint heuristic works."""
        self.fixture.given_jgex_problem("a b c = triangle a b c")
        self.fixture.when_applying_heuristic(MidpointHeuristic())
        self.fixture.then_there_should_be_new_assumptions_of_predicate(
            PredicateType.MIDPOINT
        )
        self.fixture.then_there_should_be_at_least_n_new_points(1)


class TestCentersOfCyclicHeuristic:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = HeuristicFixture()

    def test_successful_centers_of_cyclic_heuristic(self):
        """Test that centers of cyclic heuristic works."""
        self.fixture.given_jgex_problem("a b c = triangle a b c; d = on_circum d a b c")
        self.fixture.when_applying_heuristic(
            CentersOfCyclicHeuristic(), max_new_points=5
        )
        self.fixture.then_there_should_be_new_assumptions_of_predicate(
            PredicateType.CIRCUMCENTER
        )
        self.fixture.then_there_should_be_at_least_n_new_points(1)

    def test_center_after_solver_application(self):
        """Test that we find center of cyclic that were found by the solver."""
        self.fixture.given_jgex_problem(
            "a b = segment a b; c = on_dia c a b; d = on_dia d a b"
        )
        self.fixture.when_applying_heuristic(CentersOfCyclicHeuristic())
        self.fixture.then_there_should_be_at_least_n_new_points(1)
        self.fixture.then_there_should_be_new_assumptions_of_predicate(
            PredicateType.CIRCUMCENTER
        )


class TestFootHeuristic:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = HeuristicFixture()

    def test_successful_foot_heuristic(self):
        """Test that foot heuristic works."""
        self.fixture.given_jgex_problem("a b = segment a b; c = on_tline c b a b")
        self.fixture.when_applying_heuristic(FootHeuristic())
        self.fixture.then_there_should_be_at_least_n_new_points(1)


class TestCompleteThePictureHeuristics:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = HeuristicFixture()

    def test_complete_the_picture(self):
        """Test applying multiple heuristics."""
        self.fixture.given_jgex_problem(
            "a b c = triangle a b c; d e = segment d e; f = on_circum f a b c;"
            " g = on_aline0 g a b c d e f d ? eqangle f a a b f c c b"
        )
        self.fixture.when_applying_complete_the_picture_heuristics()
        self.fixture.then_there_should_be_at_least_n_new_points(2)
        self.fixture.then_heuristic_should_have_added_predicates_of_predicate(
            HeuristicName.CENTERS, PredicateType.CIRCUMCENTER
        )
        self.fixture.then_heuristic_should_have_added_predicates_of_predicate(
            HeuristicName.LINE_INTERSECTIONS, PredicateType.COLLINEAR
        )
        self.fixture.then_heuristic_should_have_added_predicates_of_predicate(
            HeuristicName.ANGLE_VERTICES, PredicateType.COLLINEAR
        )


class HeuristicFixture:
    def __init__(self):
        self.rng = np.random.default_rng(42)
        self._problem: JGEXFormulation | None = None
        self._resulting_nc_problem: ProblemSetup | None = None
        self._resulting_jgex_problem: JGEXFormulation | None = None
        self._added_clauses: list[JGEXClause] | None = None

    @property
    def problem(self) -> JGEXFormulation:
        if self._problem is None:
            raise ValueError("No problem was given")
        return self._problem

    @property
    def result(self) -> ProblemSetup:
        if self._resulting_nc_problem is None:
            raise ValueError("No heuristics applied")
        return self._resulting_nc_problem

    @property
    def points_added(self) -> int:
        return len(self.result.points) - len(self.problem.points)

    def given_jgex_problem(self, problem_txt: str):
        self._problem = JGEXFormulation.from_text(problem_txt)

    def when_applying_heuristic(
        self,
        heuristic: Heuristic,
        max_new_points: int | None = None,
    ):
        if self._problem is None:
            raise ValueError("No problem was given")

        self._initial_nc_problem = build_nc_problem_from_jgex_problem(
            self.problem, self.rng
        )
        (
            self._resulting_nc_problem,
            self._added_clauses_consequences,
        ) = apply_heuristics_on_nc_problem(
            self._initial_nc_problem, self.problem, heuristic, self.rng, max_new_points
        )

    def when_applying_complete_the_picture_heuristics(
        self, max_new_points: int | None = None
    ):
        if self._problem is None:
            raise ValueError("No problem was given")
        self._initial_nc_problem = build_nc_problem_from_jgex_problem(
            self.problem, self.rng
        )
        (
            self._resulting_nc_problem,
            self._resulting_jgex_problem,
            self._clauses_consequences_per_heuristic,
        ) = apply_complete_the_picture_heuristics(
            self._initial_nc_problem, self.problem, self.rng, max_new_points
        )

    def then_there_should_be_no_new_points(self):
        assert self.points_added == 0

    def then_there_should_be_at_least_n_new_points(self, n: int):
        assert self.points_added >= n

    def then_there_should_be_at_most_n_new_points(self, n: int):
        assert self.points_added <= n

    def then_clauses_should_be_added(self, expected_clauses: list[JGEXClause]):
        if self._added_clauses is None:
            raise ValueError("No clauses were added")
        assert self._added_clauses == expected_clauses, pretty_basemodel_list_diff(
            self._added_clauses, expected_clauses
        )

    def then_heuristic_should_have_added_predicates_of_predicate(
        self, heuristic: HeuristicName, predicate_type: PredicateType
    ):
        clauses_added_by_heuristic = self._clauses_consequences_per_heuristic[heuristic]
        added_statments_by_heuristic: list[PredicateConstruction] = []
        for clause_consequences in clauses_added_by_heuristic:
            for constru_consequence in clause_consequences.construction_consequences:
                added_statments_by_heuristic.extend(
                    constru_consequence.added_predicates
                )

        added_predicates_of_predicate = [
            predicate_construction
            for predicate_construction in added_statments_by_heuristic
            if predicate_construction.predicate_type == predicate_type
        ]

        assert len(added_predicates_of_predicate) > 0, (
            f"No predicates of predicate {predicate_type} found. "
            f"All added predicates: {added_statments_by_heuristic}"
        )

    def then_there_should_be_new_assumptions_of_predicate(
        self, predicate_type: PredicateType
    ):
        new_assumptions = [
            assumption
            for assumption in self.result.assumptions
            if assumption not in self._initial_nc_problem.assumptions
        ]
        new_assumptions_of_predicate = [
            assumption
            for assumption in new_assumptions
            if assumption.predicate_type == predicate_type
        ]
        assert len(new_assumptions_of_predicate) > 0, (
            f"No new assumptions of predicate {predicate_type} found. "
            f"All new assumptions: {new_assumptions}"
        )


def test_get_available_from_alphabet():
    """Test that get_available_from_alphabet returns the first unused point name."""
    used_points = [
        PredicateArgument("a"),
        PredicateArgument("b"),
        PredicateArgument("c"),
    ]
    available = get_available_from_alphabet(used_points)
    assert available == PredicateArgument("d")

    # Test with all points used except one
    used_points = [
        PredicateArgument("a"),
        PredicateArgument("b"),
        PredicateArgument("c"),
        PredicateArgument("e"),
    ]
    available = get_available_from_alphabet(used_points)
    assert available == PredicateArgument("d")

    # Test with empty list
    available = get_available_from_alphabet([])
    assert available == PredicateArgument("a")



from __future__ import annotations

from fractions import Fraction
from typing import TYPE_CHECKING, Literal

from numpy import pi

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Line, Point, Segment
from newclid.symbols.symbols_registry import LineOrCircle, SymbolsRegistry
from newclid.tools import fraction_to_angle, get_quotient, str_to_fraction

if TYPE_CHECKING:
    from newclid.proof_state import ProofState


def angle_between_4_points(a: Point, b: Point, c: Point, d: Point):
    return ((d.num - c.num).angle() - (b.num - a.num).angle()) % pi


class ConstantAngle(PredicateInterface):
    """aconst AB CD Y -
    Represent that the rotation needed to go from line AB to line CD,
    oriented on the clockwise direction is Y.

    The syntax of Y is either a fraction of pi like 2pi/3 for radians
    or a number followed by a 'o' like 120o for degrees.
    """

    predicate_type: Literal[PredicateType.CONSTANT_ANGLE] = PredicateType.CONSTANT_ANGLE

    line1: Line
    line2: Line
    angle: Fraction

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c, d, y = args
        if a == b or c == d:
            return None
        a, b = sorted((a, b))
        c, d = sorted((c, d))
        f = str_to_fraction(y)
        f %= 1
        return (a, b, c, d, PredicateArgument(fraction_to_angle(f)))

    def check_numerical(self) -> bool:
        a, b, c, d = self.line1 + self.line2
        angle_value = angle_between_4_points(a, b, c, d)
        expected_value = float(self.angle) * pi % pi
        return close_enough(angle_value, expected_value)

    def symbols(self, symbols: SymbolsRegistry) -> tuple[LineOrCircle, ...]:
        a, b = self.line1
        l1 = symbols.lines.line_thru_pair(a, b)
        c, d = self.line2
        l2 = symbols.lines.line_thru_pair(c, d)
        return (l1, l2)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        a, b = self.line1
        c, d = self.line2
        return (
            PredicateArgument(a.name),
            PredicateArgument(b.name),
            PredicateArgument(c.name),
            PredicateArgument(d.name),
            PredicateArgument(fraction_to_angle(self.angle)),
        )

    def __str__(self) -> str:
        a, b = self.line1
        c, d = self.line2
        return f"∠({a}{b},{c}{d}) = {fraction_to_angle(self.angle)}"


class ACompute(PredicateInterface):
    """acompute AB CD"""

    predicate_type: Literal[PredicateType.A_COMPUTE] = PredicateType.A_COMPUTE

    segment1: Segment
    segment2: Segment

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c, d = args
        if a == b or c == d:
            return None
        a, b = sorted((a, b))
        c, d = sorted((c, d))
        return (a, b, c, d)

    def check_numerical(self) -> bool:
        a, b, c, d = self.segment1 + self.segment2
        ang = angle_between_4_points(a, b, c, d)
        if close_enough(ang, pi):
            ang = 0
        get_quotient(ang / pi)
        return True

    def check(self, proof_state: ProofState) -> bool:
        aconst = aconst_from_acompute(self)
        return proof_state.check(aconst)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.segment1 + self.segment2)

    def __str__(self) -> str:
        aconst = aconst_from_acompute(self)
        return str(aconst)


def aconst_from_acompute(acompute: ACompute) -> ConstantAngle:
    a, b, c, d = acompute.segment1 + acompute.segment2
    y = get_quotient(angle_between_4_points(a, b, c, d) / pi)
    return ConstantAngle(line1=(a, b), line2=(c, d), angle=y)



"""Definitions made for specific problems."""

from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.definition import JGEXDefinition, SketchConstruction, v
from newclid.jgex.sketch import (
    sketch_2l1c,
    sketch_3peq,
    sketch_e5128,
    sketch_function_name,
    sketch_ninepoints,
    sketch_test_r20,
    sketch_test_r25,
)

TWOL1C = JGEXDefinition(
    name=JGEXConstructionName.TWOL1C,
    args=(v("x"), v("y"), v("z"), v("i"), v("a"), v("b"), v("c"), v("o")),
    rely_on_points={
        v("x"): (v("a"), v("b"), v("c"), v("o"), v("y"), v("z"), v("i")),
        v("y"): (v("a"), v("b"), v("c"), v("o"), v("x"), v("z"), v("i")),
        v("z"): (v("a"), v("b"), v("c"), v("o"), v("x"), v("y"), v("i")),
        v("i"): (v("a"), v("b"), v("c"), v("o"), v("x"), v("y"), v("z")),
    },
    requirements=JGEXClause.from_str("cong o a o b, ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "x y z i : coll x a c, coll y b c, cong o a o z, coll i o z, cong i x i y, cong i y i z, perp i x a c, perp i y b c"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_2l1c.__name__),
            args=(v("a"), v("b"), v("c"), v("o")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("o")),
    output_points=(v("x"), v("y"), v("z"), v("i")),
)


E5128 = JGEXDefinition(
    name=JGEXConstructionName.E5128,
    args=(v("x"), v("y"), v("a"), v("b"), v("c"), v("d")),
    rely_on_points={
        v("x"): (v("a"), v("b"), v("c"), v("d"), v("y")),
        v("y"): (v("a"), v("b"), v("c"), v("d"), v("x")),
    },
    requirements=JGEXClause.from_str("cong c b c d, perp b c b a")[0],
    clauses=JGEXClause.from_str(
        "x y : cong c b c x, coll y a b, coll x y d, eqangle a b a d x a x y"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_e5128.__name__),
            args=(v("a"), v("b"), v("c"), v("d")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d")),
    output_points=(v("x"), v("y")),
)

THREEPEQ = JGEXDefinition(
    name=JGEXConstructionName.THREEPEQ,
    args=(v("x"), v("y"), v("z"), v("a"), v("b"), v("c")),
    rely_on_points={
        v("z"): (v("b"), v("c")),
        v("x"): (v("a"), v("b"), v("c"), v("z"), v("y")),
        v("y"): (v("a"), v("b"), v("c"), v("z"), v("x")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "z : coll z b c ; x y : coll x a b, coll y a c, coll x y z, cong z x z y"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_3peq.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"), v("y"), v("z")),
)


TEST_R20 = JGEXDefinition(
    name=JGEXConstructionName.TEST_R20,
    args=(v("o"), v("a"), v("b"), v("c")),
    rely_on_points={
        v("b"): (v("a"), v("c")),
        v("o"): (v("a"), v("b"), v("c")),
    },
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("circle o a b c, coll o a c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_test_r20.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("o"), v("a"), v("b"), v("c")),
)


TEST_R25 = JGEXDefinition(
    name=JGEXConstructionName.TEST_R25,
    args=(v("a"), v("b"), v("c"), v("d"), v("m")),
    rely_on_points={v("m"): (v("a"), v("b"), v("c"), v("d"))},
    requirements=JGEXClause.from_str("")[0],
    clauses=JGEXClause.from_str("midp m a b, midp m c d"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_test_r25.__name__), args=()
        ),
    ),
    input_points=(),
    output_points=(v("a"), v("b"), v("c"), v("d"), v("m")),
)

PROBLEM_SPECIFIC_CONSTRUCTIONS = [
    TWOL1C,
    E5128,
    THREEPEQ,
    TEST_R20,
    TEST_R25,
]
NINEPOINTS = JGEXDefinition(
    name=JGEXConstructionName.NINEPOINTS,
    args=(v("x"), v("y"), v("z"), v("i"), v("a"), v("b"), v("c")),
    rely_on_points={
        v("x"): (v("b"), v("c")),
        v("y"): (v("c"), v("a")),
        v("z"): (v("a"), v("b")),
        v("i"): (v("x"), v("y"), v("z")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "x : coll x b c, cong x b x c; y : coll y c a, cong y c y a; z : coll z a b, cong z a z b; i : cong i x i y, cong i y i z"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_ninepoints.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"), v("y"), v("z"), v("i")),
)



from dataclasses import dataclass

from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.typing import ColorType

from newclid.animation.artists_in_animation import ArtistInAnimation, PointArtists
from newclid.draw.jgex_constructions import draw_jgex_constructions
from newclid.draw.theme import DrawTheme
from newclid.jgex.clause import JGEXClause, is_numerical_argument
from newclid.jgex.formulation import JGEXFormulation
from newclid.predicate_types import PredicateArgument
from newclid.symbols.symbols_registry import SymbolsRegistry


@dataclass
class ClauseArtists:
    clause: JGEXClause
    constructions: list[ArtistInAnimation]
    new_points: list[PointArtists]
    used_points: list[PointArtists]

    def manifest(
        self,
        aux_point_names: set[PredicateArgument],
        highlight_color: ColorType | None = None,
    ) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for used_point in self.used_points:
            updated_artists.update(used_point.highlight())
        for construction in self.constructions:
            if construction.highlight(color=highlight_color):
                updated_artists.add(construction.artist)
        for point in self.new_points:
            color = highlight_color if point.name not in aux_point_names else None
            updated_artists.update(point.highlight(color=color))
        return updated_artists

    def gray_out(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for point in self.used_points:
            updated_artists.update(point.gray_out())
        for construction in self.constructions:
            if construction.gray_out():
                updated_artists.add(construction.artist)
        for point in self.new_points:
            updated_artists.update(point.gray_out())
        return updated_artists

    def hide(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for construction in self.constructions:
            if construction.hide():
                updated_artists.add(construction.artist)
        return updated_artists


def init_clauses_to_draw(
    ax: Axes,
    jgex_problem: JGEXFormulation,
    point_name_to_artist: dict[str, PointArtists],
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> dict[JGEXClause, ClauseArtists]:
    clauses_to_draw: dict[JGEXClause, ClauseArtists] = {}
    for clause in jgex_problem.clauses:
        new_points_artists: list[PointArtists] = [
            point_name_to_artist[pname] for pname in clause.points
        ]
        used_points_in_clause_artists: list[PointArtists] = [
            point_name_to_artist[pname]
            for construction in clause.constructions
            for pname in construction.args
            if not is_numerical_argument(pname)
        ]
        construction_artists: list[ArtistInAnimation] = []
        for construction in clause.constructions:
            for artist in draw_jgex_constructions(
                ax, construction, symbols_registry=symbols, theme=theme
            ):
                construction_artists.append(ArtistInAnimation(artist))
        clauses_to_draw[clause] = ClauseArtists(
            clause=clause,
            new_points=new_points_artists,
            used_points=used_points_in_clause_artists,
            constructions=construction_artists,
        )
    return clauses_to_draw



import logging
import os
import sys
from pathlib import Path

import numpy as np

from newclid.agent.agent_builder import make_agent
from newclid.api import GeometricSolverBuilder
from newclid.cli import parse_cli_args
from newclid.jgex.problem_builder import JGEXProblemBuilder


def find_ggb_files(directory: Path):
    for entry in os.listdir(directory):
        if entry.endswith(".ggb"):
            yield entry


LOGGER = logging.getLogger(__name__)


def main() -> None:
    options, problem_builder = parse_cli_args(sys.argv[1:])
    logging.basicConfig(level=options.log_level.value)

    rng = np.random.default_rng(options.seed)
    solver_builder = GeometricSolverBuilder(rng)

    if options.agent is not None:
        agent = make_agent(options.agent)
        solver_builder.with_deductive_agent(agent)

    solver = solver_builder.build(problem_builder.build())

    jgex_problem = None
    if isinstance(problem_builder, JGEXProblemBuilder):
        jgex_problem = problem_builder.jgex_problem

    if options.output_dir:
        solver.draw_figure(
            out_file=options.output_dir / "initial_figure.svg",
            jgex_problem=jgex_problem,
        )

    solver.run()
    if solver.run_infos is not None:
        LOGGER.info(f"Run infos: {solver.run_infos.model_dump_json(indent=2)}")

    if options.output_dir is not None:
        solver.write_all_outputs(options.output_dir, jgex_problem=jgex_problem)


if __name__ == "__main__":
    main()
